"""Support for Subversion repositories."""

from __future__ import annotations

import logging
import os
import re
import weakref
from enum import IntEnum
from importlib import import_module
from typing import (Any, Dict, Final, Mapping, Optional, Sequence,
                    TYPE_CHECKING, Union, cast)
from urllib.parse import urlparse

from django.conf import settings
from django.utils.encoding import force_bytes, force_str
from django.utils.translation import gettext as _
from typing_extensions import TypedDict

from reviewboard import get_manual_url
from reviewboard.admin.server import get_data_dir
from reviewboard.diffviewer.parser import DiffParser
from reviewboard.scmtools.certs import Certificate
from reviewboard.scmtools.core import (
    Branch,
    Commit,
    HEAD,
    PRE_CREATION,
    Revision,
    SCMTool,
    UNKNOWN,
)
from reviewboard.scmtools.errors import (AuthenticationError,
                                         RepositoryNotFoundError,
                                         SCMError,
                                         UnverifiedCertificateError)
from reviewboard.scmtools.forms import StandardSCMToolRepositoryForm
from reviewboard.scmtools.svn.utils import (collapse_svn_keywords,
                                            has_expanded_svn_keywords)
from reviewboard.ssh import utils as sshutils

if TYPE_CHECKING:
    from reviewboard.diffviewer.parser import ParsedDiffFile
    from reviewboard.scmtools.core import RevisionID
    from reviewboard.scmtools.models import Repository
    from reviewboard.scmtools.svn.base import (
        Client,
        SVNDirEntry,
        SVNLogEntry,
        SVNRepositoryInfoDict,
    )


logger = logging.getLogger(__name__)


# These will be set later in recompute_svn_backend().
_SVNClientBackend: Optional[type[Client]] = None
has_svn_backend: bool = False


# Revision atom for use with IntelliJ's empty revision labels.
_IDEA_EMPTY = Revision('IDEA_EMPTY')


# Register these URI schemes so we can handle them properly.
sshutils.ssh_uri_schemes.append('svn+ssh')

sshutils.register_rbssh('SVN_SSH')


# Compiled regexes for parsing diffs.
_revision_re: Optional[re.Pattern] = None
_working_copy_re: Optional[re.Pattern] = None
_binary_revision_re: Optional[re.Pattern] = None


class SVNCertificateFailures(IntEnum):
    """SVN HTTPS certificate failure codes.

    These map to the various SVN HTTPS certificate failures in libsvn.
    """

    #: The certificate is not yet valid.
    NOT_YET_VALID = 1 << 0

    #: The certificate has expired.
    EXPIRED = 1 << 1

    #: The certificate does not match the hostname.
    CN_MISMATCH = 1 << 2

    #: The certificate is self-signed or signed by an untrusted authority.
    UNKNOWN_CA = 1 << 3


class RawSSLTrustDict(TypedDict):
    """A dictionary of SSL trust data to verify.

    For details, see
    https://pysvn.sourceforge.io/Docs/pysvn_prog_ref.html#pysvn_client_callback_ssl_server_trust_prompt

    Version Added:
        6.0
    """

    #: The bitmask of failures:
    #:
    #: Type:
    #:     int
    failures: int

    #: The SSL certificate fingerprint.
    finger_print: str

    #: The hostname that served the certificate.
    #:
    #: Type:
    #:     str
    hostname: str

    #: The issuer of the certificate.
    #:
    #: Type:
    #:     str
    issuer_dname: str

    #: The realm serving the certificate.
    #:
    #: Type:
    #:     str
    realm: str

    #: The start of the certificate validity period.
    #:
    #: This is in :term:`ISO8601 format`.
    #:
    #: Type:
    #:     str
    valid_from: str

    #: The end of the certificate validity period.
    #:
    #: This is in :term:`ISO8601 format`.
    #:
    #: Type:
    #:     str
    valid_until: str


class SVNRepositoryForm(StandardSCMToolRepositoryForm):
    """Form for editing SVN repositories.

    Version Added:
        6.0
    """

    def clean(self) -> dict[str, Any]:
        """Perform validation on the form.

        Returns:
            dict:
            The cleaned form data.
        """
        cleaned_data = super().clean()
        assert cleaned_data is not None

        path = cleaned_data.get('path', '')

        if path:
            url_parts = urlparse(path)

            if not url_parts.scheme:
                self.add_error('path', _(
                    'The path to the SVN repository must be a URL. To specify '
                    'a local repository, use a file:// URL.'))

        return cleaned_data


class SVNTool(SCMTool):
    """Repository support for Subversion.

    Subversion is an open source centralized source code management system
    maintained by the Apache Software Foundation. It supports tagging and
    branching, atomic commits, full remote communication, file/directory-level
    metadata, large binary files, and more.

    Review Board uses Subversion's native diff format. This has had some
    changes and inconsistencies over time, but is otherwise a fairly
    comprehensive diff format.
    """

    scmtool_id = 'subversion'
    name = 'Subversion'
    supports_post_commit = True
    dependencies = {
        # This will get filled in later in recompute_svn_backend()
        'modules': [],
    }
    repository_form = SVNRepositoryForm

    #: The number of commits to retrieve per page.
    COMMITS_PAGE_LIMIT: Final[int] = 31

    ######################
    # Instance variables #
    ######################

    #: The configured Subversion client backend.
    #:
    #: Type:
    #:     reviewboard.scmtools.svn.base.Client
    client: Client

    def __init__(
        self,
        repository: Repository,
    ) -> None:
        """Initialize the Subversion support.

        Args:
            repository (reviewboard.scmtools.models.Repository):
                The Subversion repository that this tool will communicate with.
        """
        repopath = repository.path

        if repopath.endswith('/'):
            repopath = repopath.rstrip('/')

        super().__init__(repository=repository)

        if repository.local_site:
            local_site_name = repository.local_site.name
        else:
            local_site_name = None

        credentials = repository.get_credentials()

        client = self.build_client(repo_path=repopath,
                                   username=credentials['username'],
                                   password=credentials['password'],
                                   local_site_name=local_site_name)
        self.client = client

        # If we assign a function to the pysvn Client that accesses anything
        # bound to SVNClient, it'll end up keeping a reference and a copy of
        # the function for every instance that gets created, and will never
        # let go. This will cause a rather large memory leak.
        #
        # The solution is to access a weakref instead. The weakref will
        # reference the repository, but it will safely go away when needed.
        # The function we pass can access that without causing the leaks
        repository_ref = weakref.ref(repository)
        client.set_ssl_server_trust_prompt(
            lambda trust_dict:
            SVNTool._ssl_server_trust_prompt(trust_dict, repository_ref()))

    def get_file(
        self,
        path: str,
        revision: RevisionID = HEAD,
        *args,
        **kwargs,
    ) -> bytes:
        """Return the contents of a file from a repository.

        This attempts to return the raw binary contents of a file from the
        repository, given a file path and revision.

        Args:
            path (str):
                The path to the file in the repository.

            revision (reviewboard.scmtools.core.Revision, optional):
                The revision to fetch. Subclasses should default this to
                :py:data:`HEAD`.

            *args (tuple, unused):
                Additional unused positional arguments.

            **kwargs (dict, unused):
                Additional unused keyword arguments.

        Returns:
            bytes:
            The returned file contents.

        Raises:
            reviewboard.scmtools.errors.FileNotFoundError:
                The file could not be found in the repository.

            reviewboard.scmtools.errors.InvalidRevisionFormatError:
                The ``revision`` argument was in an invalid format.

            reviewboard.scmtools.errors.SCMError:
                An unexpected error was encountered with the repository.
        """
        return self.client.get_file(path, revision)

    def get_branches(self) -> Sequence[Branch]:
        """Return a list of all branches on the repository.

        This will fetch a list of all known branches for use in the API and
        New Review Request page.

        This assumes the standard layout in the repository.

        Returns:
            list of reviewboard.scmtools.core.Branch:
            The list of branches in the repository. One (and only one) will
            be marked as the default branch.

        Raises:
            reviewboard.scmtools.errors.SCMError:
                The repository tool encountered an error.
        """
        results: list[Branch] = []

        root_dirents = self.client.list_dir('/')
        default = True

        if 'trunk' in root_dirents:
            # Looks like the standard layout. Adds trunk and any branches.
            trunk = root_dirents['trunk']
            results.append(self._create_branch_from_dirent(
                name='trunk',
                dirent=trunk,
                default=True))
            default = False

        if 'branches' in root_dirents:
            dirents = self.client.list_dir('branches')
            results += [
                self._create_branch_from_dirent(name=name,
                                                dirent=dirents[name])
                for name in sorted(dirents.keys())
            ]

        # Add anything else from the root of the repository. This is a
        # catch-all for repositories which do not use the standard layout, and
        # for those that do, will include any additional top-level directories
        # that people may have.
        for name in sorted(root_dirents.keys()):
            if name not in ('trunk', 'branches'):
                results.append(self._create_branch_from_dirent(
                    name=name,
                    dirent=root_dirents[name],
                    default=default))
                default = False

        return results

    def get_commits(
        self,
        branch: Optional[str] = None,
        start: Optional[str] = None,
    ) -> Sequence[Commit]:
        """Return a list of commits backward in history from a given point.

        This will fetch a batch of commits from the repository for use in the
        API and New Review Request page.

        The resulting commits will be in order from newest to oldest, and
        should return up to a fixed number of commits (usually 30, but this
        depends on the type of repository and its limitations). It may also be
        limited to commits that exist on a given branch (if supported by the
        repository).

        This can be called multiple times in succession using the
        :py:attr:`Commit.parent` of the last entry as the ``start`` parameter
        in order to paginate through the history of commits in the repository.

        Args:
            branch (str, optional):
                The branch to limit commits to. This may not be supported by
                all repositories.

            start (str, optional):
                The commit to start at. If not provided, this will fetch the
                first commit in the repository.

        Returns:
            list of reviewboard.scmtools.core.Commit:
            The list of commits, in order from newest to oldest.

        Raises:
            reviewboard.scmtools.errors.SCMError:
                The repository tool encountered an error.

            NotImplementedError:
                Commits retrieval is not available for this type of repository.
        """
        commits = self.client.get_log(branch or '/',
                                      start=start,
                                      limit=self.COMMITS_PAGE_LIMIT,
                                      limit_to_path=False)

        results: list[Commit] = []

        # We fetch one more commit than we care about, because the entries in
        # the svn log doesn't include the parent revision.
        for i in range(len(commits) - 1):
            commit = commits[i]
            parent = commits[i + 1]
            results.append(self._build_commit(data=commit,
                                              parent=parent['revision']))

        # If there were fewer than the requested number of commits fetched,
        # also include the last one in the list so we don't leave off the
        # initial revision.
        if len(commits) < self.COMMITS_PAGE_LIMIT:
            commit = commits[-1]
            results.append(self._build_commit(data=commit))

        return results

    def get_change(
        self,
        revision: str,
    ) -> Commit:
        """Return an individual commit with the given revision.

        This will fetch information on the given commit, if found, including
        its commit message and list of modified files.

        Args:
            revision (str):
                The revision of the commit.

        Returns:
            Commit:
            The resulting commit with the given revision.

        Raises:
            reviewboard.scmtools.errors.SCMError:
                Error retrieving information on this commit.
        """
        commits = self.client.get_log('/', start=revision, limit=2)

        if len(commits) > 1:
            base_revision = commits[1]['revision']
        else:
            base_revision = '0'

        diff = self.client.diff(base_revision, revision)

        commit = self._build_commit(data=commits[0],
                                    parent=base_revision)
        commit.diff = diff

        return commit

    def normalize_patch(
        self,
        patch: bytes,
        filename: str,
        revision: str,
    ) -> bytes:
        """Normalize a diff/patch file before it's applied.

        This will check the patch for any ``$...$`` keywords. If found, the
        repository will be queried for matches, and those will be collapsed
        in the diff. This ensures that the file can be applied on top of a
        normalized source file, and compared to other normalized patches.

        Args:
            patch (bytes):
                The diff/patch file to normalize.

            filename (str):
                The name of the file being changed in the diff.

            revision (str):
                The revision of the file being changed in the diff.

        Returns:
            bytes:
            The resulting diff/patch file.
        """
        if revision != PRE_CREATION and has_expanded_svn_keywords(patch):
            keywords = self.client.get_keywords(filename, revision)

            if keywords:
                return collapse_svn_keywords(patch, force_bytes(keywords))

        return patch

    def parse_diff_revision(
        self,
        filename: bytes,
        revision: bytes,
        *args,
        **kwargs,
    ) -> tuple[bytes, Union[bytes, Revision]]:
        """Parse and return a filename and revision from a diff.

        Args:
            filename (bytes):
                The filename as represented in the diff.

            revision (bytes):
                The revision as represented in the diff.

            *args (tuple, unused):
                Unused positional arguments.

            **kwargs (dict, unused):
                Unused keyword arguments.

        Returns:
            tuple:
            A tuple containing two items:

            Tuple:
                0 (bytes):
                    The normalized filename.

                1 (bytes or reviewboard.scmtools.core.Revision):
                    The normalized revision.
        """
        # Revision parsing now happens inside SVNDiffParser.
        return filename, revision

    def get_repository_info(self) -> SVNRepositoryInfoDict:
        """Return information on the repository.

        Returns:
            dict:
            A dictionary containing information on the repository.

            See :py:class:`reviewboard.scmtools.svn.base.SVNRepositoryInfoDict`
            for contents.

        Raises:
            NotImplementedError:
                Repository information retrieval is not implemented by this
                type of repository. Callers should specifically check for this,
                as it's considered a valid result.
        """
        return self.client.repository_info

    def get_parser(
        self,
        data: bytes,
    ) -> SVNDiffParser:
        """Return a diff parser used to parse diff data.

        The diff parser will be responsible for parsing the contents of the
        diff, and should expect (but validate) that the diff content is
        appropriate for the type of repository.

        Subclasses should override this.

        Args:
            data (bytes):
                The diff data to parse.

        Returns:
            SVNDiffParser:
            The diff parser used to parse this data.
        """
        return SVNDiffParser(data)

    def _create_branch_from_dirent(
        self,
        *,
        name: str,
        dirent: SVNDirEntry,
        default: bool = False,
    ) -> Branch:
        """Return a Branch object from a Subversion directory entry.

        Args:
            name (str):
                The name of the directory entry.

            dirent (reviewboard.scmtools.svn.base.SVNDirEntry):
                The directory entry to parse.

            default (bool, optional):
                Whether this is the default branch.

        Returns:
            reviewboard.scmtools.core.Branch:
            The resulting branch.
        """
        return Branch(
            id=dirent['path'].strip('/'),
            name=name,
            commit=dirent['created_rev'],
            default=default)

    def _build_commit(
        self,
        *,
        data: SVNLogEntry,
        parent: Union[bytes, str] = '',
    ) -> Commit:
        """Return a Commit object from the provided data.

        Args:
            data (reviewboard.scmtools.svn.base.SVNLogEntry):
                The log entry to build the data from.

            parent (bytes or str, optional):
                The parent commit revision.

        Returns:
            reviewboard.scmtools.core.Commit:
            The resulting commit.
        """
        date = data.get('date')
        norm_date: str = ''

        if date:
            norm_date = str(date.isoformat())

        return Commit(
            author_name=force_str(data.get('author', ''), errors='replace'),
            id=force_str(data['revision']),
            date=norm_date,
            message=force_str(data.get('message', ''), errors='replace'),
            parent=force_str(parent))

    @classmethod
    def _ssl_server_trust_prompt(
        cls,
        trust_data: RawSSLTrustDict,
        repository: Optional[Repository],
    ) -> tuple[bool, int, bool]:
        """Callback for SSL cert verification.

        This will be called when accessing a repository with an SSL cert.
        We will look up a matching cert in the database and see if it's
        accepted.

        Args:
            trust_data (RawSSLTrustDict):
                The Subversion SSL trust data.

            repository (reviewboard.scmtools.models.Repository):
                The repository triggering the SSL cert verification.

        Returns:
            tuple:
            A 3-tuple containing:

            Tuple:
                0 (bool):
                    Whether the certificate is verified.

                1 (int):
                    A bitmask of Subversion SSL trust failure codes.

                2 (bool):
                    Whether to store the trust info within the filesystem.

                    This is always ``False``, since we want to handle this
                    on our end.
        """
        if repository is None:
            return False, 0, False

        saved_cert = repository.extra_data.get('cert', {})
        cert = cast(Dict[str, Any], trust_data.copy())
        del cert['failures']

        return saved_cert == cert, trust_data['failures'], False

    @staticmethod
    def on_ssl_failure(
        client: Client,
        e: Exception,
        path: str,
        cert_data: dict[str, Any],
    ) -> None:
        """Handle a SSL certificate failure.

        This will determine if the error contains unverified SSL certificate
        information that needs to be returned. If found, a suitable error
        will be generated and raised.

        Args:
            client (reviewboard.scmtools.svn.base.Client):
                The client talking to Subversion.

            e (Exception):
                The raw exception found during SSL communication.

            path (str):
                The repository path.

            cert_data (RawSSLTrustDict):
                The dictionary containing Subversion SSL certificate
                information.

        Raises:
            reviewboard.scmtools.errors.AuthenticationError:
                There was an error authenticating, rather than an issue with
                the SSL certificate.

            reviewboard.scmtools.errors.RepositoryNotFoundError:
                The repository was not found.

            reviewboard.scmtools.errors.UnverifiedCertificateError:
                The certificate was not verified.

                Details are in the error message and associated certificate.
        """
        logger.error('SVN: Failed to get repository information for %s: %s',
                     path, e)

        error = client.normalize_error(e)

        if isinstance(error, AuthenticationError):
            raise error

        if cert_data:
            failures = cert_data['failures']

            reasons: list[str] = []

            if failures & SVNCertificateFailures.NOT_YET_VALID:
                reasons.append(_('The certificate is not yet valid.'))

            if failures & SVNCertificateFailures.EXPIRED:
                reasons.append(_('The certificate has expired.'))

            if failures & SVNCertificateFailures.CN_MISMATCH:
                reasons.append(_('The certificate hostname does not '
                                 'match.'))

            if failures & SVNCertificateFailures.UNKNOWN_CA:
                reasons.append(_('The certificate is not issued by a '
                                 'trusted authority. Use the fingerprint '
                                 'to validate the certificate manually.'))

            raise UnverifiedCertificateError(
                Certificate(valid_from=cert_data['valid_from'],
                            valid_until=cert_data['valid_until'],
                            hostname=cert_data['hostname'],
                            realm=cert_data['realm'],
                            fingerprint=cert_data['finger_print'],
                            issuer=cert_data['issuer_dname'],
                            failures=reasons))

        raise RepositoryNotFoundError()

    @classmethod
    def check_repository(
        cls,
        path: str,
        username: Optional[str] = None,
        password: Optional[str] = None,
        local_site_name: Optional[str] = None,
        *args,
        **kwargs,
    ) -> None:
        """Check a repository configuration for validity.

        This should check if a repository exists and can be connected to.
        This will also check if the repository requires an HTTPS certificate.

        A failed result is returned as an exception. The exception may contain
        extra information, such as a human-readable description of the problem.
        If the repository is valid and can be connected to, no exception will
        be thrown.

        Args:
            path (str):
                The repository path.

            username (str, optional):
                The optional username for the repository.

            password (str, optional):
                The optional password for the repository.

            local_site_name (str, optional):
                The name of the Local Site that owns this repository.

            *args (tuple, unused):
                Additional unused positional arguments.

            **kwargs (dict, unused):
                Additional unused keyword arguments.

        Raises:
            reviewboard.scmtools.errors.AuthenticationError:
                The provided username/password or the configured SSH key could
                not be used to authenticate with the repository.

            reviewboard.scmtools.errors.RepositoryNotFoundError:
                A repository could not be found at the given path.

            reviewboard.scmtools.errors.SCMError:
                There was a generic error with the repository or its
                configuration.  Details will be provided in the error message.

            reviewboard.ssh.errors.BadHostKeyError:
                An SSH path was provided, but the host key for the repository
                did not match the expected key.

            reviewboard.ssh.errors.SSHError:
                An SSH path was provided, but there was an error establishing
                the SSH connection.

            reviewboard.ssh.errors.SSHInvalidPortError:
                An SSH path was provided, but the port specified was not a
                valid number.

            Exception:
                An unexpected exception has occurred. Callers should check
                for this and handle it.
        """
        super().check_repository(
            path=path,
            username=username,
            password=password,
            local_site_name=local_site_name,
            **kwargs)

        if path.startswith('https://'):
            client = cls.build_client(repo_path=path,
                                      username=username,
                                      password=password,
                                      local_site_name=local_site_name)
            client.accept_ssl_certificate(path,
                                          on_failure=cls.on_ssl_failure)

    @classmethod
    def accept_certificate(
        cls,
        path: str,
        username: Optional[str] = None,
        password: Optional[str] = None,
        local_site_name: Optional[str] = None,
        certificate: Optional[Certificate] = None,
    ) -> Mapping[str, Any]:
        """Accept the HTTPS certificate for the given repository path.

        Args:
            path (str):
                The repository path.

            username (str, optional):
                The username provided for the repository.

            password (str, optional):
                The password provided for the repository.

            local_site_name (str, optional):
                The name of the Local Site used for the repository, if any.

            certificate (reviewboard.scmtools.certs.Certificate):
                The certificate to accept.

        Returns:
            dict:
            Serialized information on the certificate.

        Raises:
            reviewboard.scmtools.errors.SCMError:
                There was an error accepting the certificate.
        """
        client = cls.build_client(repo_path=path,
                                  username=username,
                                  password=password,
                                  local_site_name=local_site_name)

        return client.accept_ssl_certificate(path)

    @classmethod
    def build_client(
        cls,
        *,
        repo_path: str,
        username: Optional[str] = None,
        password: Optional[str] = None,
        local_site_name: Optional[str] = None,
    ) -> Client:
        """Return a new Subversion client.

        This will determine and create a Subversion directory for the global
        site or Local Site, and then return a configured client.

        Args:
            repo_path (str):
                The URI to the Subversion repository.

            username (str, optional):
                The username used to authenticate with the repository.

            password (str, optional):
                The password used to authenticate with the repository.

            local_site_name (str, optional):
                The name of the Local Site that owns the repository.

        Returns:
            tuple:
            A 2-tuple of:

            Tuple:
                0 (str):
                    The Subversion configuration directory.

                1 (reviewboard.scmtools.svn.base.Client):
                    The resulting client instance.
        """
        if not has_svn_backend:
            raise ImportError(
                _('SVN integration requires PySVN. See %(url)s for '
                  'installation instructions.')
                % {
                    'url': (
                        '%sadmin/installation/linux/#subversion'
                        % get_manual_url()
                    ),
                })

        config_dir = os.path.join(get_data_dir(), '.subversion')

        if local_site_name:
            # LocalSites can have their own Subversion config, used for
            # per-LocalSite SSH keys.
            config_dir = cls._prepare_local_site_config_dir(local_site_name)
        elif not os.path.exists(config_dir):
            cls._create_subversion_dir(config_dir)

        assert _SVNClientBackend is not None
        return _SVNClientBackend(config_dir=config_dir,
                                 repopath=repo_path,
                                 username=username,
                                 password=password)

    @classmethod
    def _create_subversion_dir(
        cls,
        config_dir: str,
    ) -> None:
        """Create a local Subversion directory for data storage.

        Args:
            config_dir (str):
                The absolute path to the new Subversion directory.

        Raises:
            IOError:
                There was an error creating the directory.

                Details are in the error message.
        """
        try:
            os.mkdir(config_dir, 0o700)
        except OSError:
            raise OSError(
                _("Unable to create directory %(dirname)s, which is needed "
                  "for the Subversion configuration. Create this directory "
                  "and set the web server's user as the the owner.")
                % {'dirname': config_dir})

    @classmethod
    def _prepare_local_site_config_dir(
        cls,
        local_site_name: str,
    ) -> str:
        """Prepare a Subversion configuration for a Local Site.

        This will create the directory if it doesn't exist, and then
        prepare a Subversion configuration for linking Review Board up with
        RBSSH.

        Args:
            local_site_name (str):
                The name of the Local Site.

        Returns:
            str:
            The resulting configuration directory.
        """
        config_dir = os.path.join(get_data_dir(), '.subversion')

        if not os.path.exists(config_dir):
            cls._create_subversion_dir(config_dir)

        config_dir = os.path.join(config_dir, local_site_name)

        if not os.path.exists(config_dir):
            cls._create_subversion_dir(config_dir)

            with open(os.path.join(config_dir, 'config'), 'w') as fp:
                fp.write('[tunnels]\n')
                fp.write('ssh = rbssh --rb-local-site=%s\n' % local_site_name)

        return config_dir


class SVNDiffParser(DiffParser):
    """Diff parser for Subversion diff files.

    This is designed to be compatible with all variations of the Subversion
    diff format, supporting standard file modifications and property changes.
    """

    BINARY_STRING = b'Cannot display: file marked as a binary type.'

    def __init__(
        self,
        data: bytes,
        **kwargs,
    ) -> None:
        """Initialize the parser.

        Version Added:
            7.0

        Args:
            data (bytes):
                The diff content to parse.

            **kwargs (dict):
                Keyword arguments to pass to the parent class.

        Raises:
            TypeError:
                The provided ``data`` argument was not a ``bytes`` type.
        """
        super().__init__(data, **kwargs)

        revision_re_part = r'''
            (                      # - svn 1.9 nonexistent revision indicator
             nonexistent|          # English
             nicht\ existent|      # German
             不存在的|             # Simplified Chinese
             (?:
               revisão|            # Brazilian Portuguese
               [Rr]ev(?:ision)?|   # English, German

                                   # - svnlook uses 'rev 0' while svn diff
                                   #   uses 'revision 0'
               révision|           # French
               revisione|          # Italian
               リビジョン|         # Japanese
               리비전|             # Korean
               revisjon|           # Norwegian
               wersja|             # Polish
               版本|               # Simplified Chinese
               revisión:           # Spanish
             )\ (\d+)              # - the revision number
            )
        '''.strip()

        working_copy_re_part = r'''
            (
             cópia\ de\ trabalho|   # Brazilian Portuguese
             working\ copy|         # English
             copie\ de\ travail|    # French
             Arbeitskopie|          # German
             copia\ locale|         # Italian
             作業コピー|            # Japanese
             작업\ 사본|            # Korean
             arbeidskopi|           # Norweigan
             kopia\ robocza|        # Polish
             工作副本|              # Simplified Chinese
             copia\ de\ trabajo     # Spanish
            )
        '''.strip()

        global _revision_re
        global _working_copy_re
        global _binary_revision_re

        if _revision_re is None:
            # 'svn diff' produces patches which have the revision string
            # localized to their system locale. This is a little ridiculous,
            # but we have to deal with it because not everyone uses RBTools.
            #
            # svnlook diff creates lines like
            # '2016-05-12 12:30:05 UTC (rev 1234)'
            #
            # whereas svn diff creates lines like
            # '(Revision 94754)'
            # '        (.../branches/product-2.0) (revision 321)'
            # '        (.../trunk)     (nonexistent)'
            #
            # So we need to form a regex to match relocation information and
            # the revision number. Subversion >=1.9 adds the 'nonexistent'
            # revision string.
            _revision_re = re.compile(
                fr'''
                    ^(\(([^\)]+)\)\s)?      # creating diffs between two
                                            # branches of a remote repository
                                            # will insert extra "relocation
                                            # information" into the diff.

                    (?:\d+-\d+-\d+\ +       # svnlook-style diffs contain a
                       \d+:\d+:\d+\ +       # timestamp on each line before the
                                            # revision number.
                       [A-Z]+\ +)?          # This here is probably a really
                                            # crappy to express that, but oh
                                            # well.

                    \ *\({revision_re_part}\)$
                '''.encode('utf-8'),
                re.VERBOSE)

            # 'svn diff' also localises the (working copy) string to the system
            # locale.
            _working_copy_re = re.compile(
                fr'^\((?:{working_copy_re_part})\)$'.encode('utf-8'),
                re.VERBOSE)

            _binary_revision_re = re.compile(
                fr'(\(({revision_re_part}|{working_copy_re_part})\))'
                .encode('utf-8'),
                re.VERBOSE)

    def parse_diff_header(
        self,
        linenum: int,
        parsed_file: ParsedDiffFile,
    ) -> int:
        """Parse a standard header before changes made to a file.

        This builds upon the standard behavior to see if the change is to a
        property, rather than a file. Property changes are skipped.

        Args:
            linenum (int):
                The line number to begin parsing.

            parsed_file (reviewboard.diffviewer.parser.ParsedDiffFile):
                The file currently being parsed.

        Returns:
            int:
            The next line number to parse.

        Raises:
            reviewboard.diffviewer.errors.DiffParserError:
                There was an error parsing the diff header. This may be a
                corrupted diff, or an error in the parsing implementation.
                Details are in the error message.
        """
        lines = self.lines

        # We're looking for a SVN property change for SVN < 1.7.
        #
        # There's going to be at least 5 lines left:
        # 1) --- (blah)
        # 2) +++ (blah)
        # 3) Property changes on: <path>
        # 4) -----------------------------------------------------
        # 5) Modified: <propname>
        try:
            is_property_change = (
                lines[linenum].startswith(b'--- (') and
                lines[linenum + 1].startswith(b'+++ (') and
                lines[linenum + 2].startswith(b'Property changes on:')
            )
        except IndexError:
            is_property_change = False

        if is_property_change:
            # Subversion diffs with property changes have no really
            # parsable format. The content of a property can easily mimic
            # the property change headers. So we can't rely upon it, and
            # can't easily display it. Instead, skip it, so it at least
            # won't break diffs.
            parsed_file.skip = True
            linenum += 4

            return linenum

        # Handle deleted empty files.
        if (parsed_file.index_header_value and
            parsed_file.index_header_value.endswith(b'\t(deleted)')):
            parsed_file.deleted = True

        linenum = super().parse_diff_header(linenum, parsed_file)

        # Handle IntelliJ/IDEA empty revisions.
        if parsed_file.orig_file_details is _IDEA_EMPTY:
            parsed_file.orig_file_details = PRE_CREATION

        if parsed_file.modified_file_details is _IDEA_EMPTY:
            parsed_file.modified_file_details = HEAD

        # Handle "(nonexistent)" revisions on deleted files. This only applies
        # if the original revision is not also PRE_CREATION (which indicates an
        # added empty file).
        if (parsed_file.modified_file_details is PRE_CREATION and
            parsed_file.orig_file_details is not PRE_CREATION):
            parsed_file.deleted = True

        return linenum

    def parse_filename_header(
        self,
        s: bytes,
        linenum: int,
    ) -> tuple[bytes, Union[bytes, Revision]]:
        """Parse a filename header in the diff.

        Args:
            s (bytes):
                The filename header to parse.

            linenum (int):
                The current line number.

        Returns:
            tuple:
            A tuple containing:

            Tuple:
                0 (bytes):
                    The filename.

                1 (bytes or reviewboard.scmtools.core.Revision):
                    The revision information.
        """
        filename, revision = super().parse_filename_header(s, linenum)

        return self.parse_diff_revision(filename, revision)

    def parse_diff_revision(
        self,
        filename: Optional[bytes],
        revision: Optional[bytes],
    ) -> tuple[bytes, Union[bytes, Revision]]:
        """Parse and return a filename and revision from the diff.

        Args:
            filename (bytes):
                The filename as represented in the diff.

            revision (bytes):
                The revision as represented in the diff.

        Returns:
            tuple:
            A 2-tuple containing:

            Tuple:
                0 (bytes):
                    The processed filename.

                1 (bytes or reviewboard.scmtools.core.Revision):
                    The processed revision.

        Raises:
            reviewboard.scmtools.errors.SCMError:
                The revision could not be parsed.
        """
        if revision is None:
            revision = b''

        assert filename is not None
        assert revision is not None

        # Some diffs have additional tabs between the parts of the file
        # revisions
        revision = revision.strip()

        if _working_copy_re.match(revision):
            return filename, HEAD

        # "(revision )" is generated by a few weird tools (like IntelliJ). If
        # in the +++ line of the diff, it means HEAD, and in the --- line, it
        # means PRE_CREATION. Since the more important use case is parsing the
        # source revision, we treat it as a new file. See bugs 1937 and 2632.
        # We'll convert this to the correct meaning later in parse_diff_header.
        if revision == b'(revision )':
            return filename, _IDEA_EMPTY

        # Binary diffs don't provide revision information, so we set a fake
        # "(unknown)" in the SVNDiffParser. This will never actually appear
        # in SVN diffs.
        if revision == b'(unknown)':
            return filename, UNKNOWN

        m = _revision_re.match(revision)

        if not m:
            raise SCMError('Unable to parse diff revision header "%s"'
                           % revision.decode('utf-8'))

        # group(3) holds the revision string in braces, like '(revision 4)'
        # group(4) only matches the revision number, which might by None when
        # 'nonexistent' is given as the revision string
        norm_revision: Union[bytes, Revision, None] = m.group(4)

        if norm_revision in (None, b'0'):
            norm_revision = PRE_CREATION

        relocated_file = m.group(2)

        if relocated_file:
            if not relocated_file.startswith(b'...'):
                raise SCMError('Unable to parse SVN relocated path "%s"'
                               % relocated_file.decode('utf-8'))

            filename = b'%s/%s' % (relocated_file[4:], filename)

        return filename, norm_revision

    def parse_special_header(
        self,
        linenum: int,
        parsed_file: ParsedDiffFile,
    ) -> int:
        """Parse a special diff header marking the start of a new file's info.

        This will look for:

        * An empty ``Index:`` line, which suggests a change to a property
          rather than a file (in older versions of SVN diffs).
        * A populated ``Index:`` line, which is required for any changes to
          files (and may be present for property changes in newer versions of
          SVN diffs).
        * Changes to binary files, which lack revision information and must
          be populated with defaults.

        Args:
            linenum (int):
                The line number to begin parsing.

            parsed_file (reviewboard.diffviewer.parser.ParsedDiffFile):
                The file currently being parsed.

        Returns:
            int:
            The next line number to parse.

        Raises:
            reviewboard.diffviewer.errors.DiffParserError:
                There was an error parsing the special header. This may be
                a corrupted diff, or an error in the parsing implementation.
                Details are in the error message.
        """
        lines = self.lines

        if linenum + 1 < len(lines) and lines[linenum] == b'Index:':
            # This is an empty Index: line. This might mean we're parsing
            # a property change.
            return linenum + 2

        linenum = super().parse_special_header(
            linenum, parsed_file)

        file_index = parsed_file.index_header_value

        if file_index is None:
            return linenum

        try:
            line = lines[linenum]
        except IndexError:
            return linenum

        # If the diff has been generated with --diff-cmd=diff --force,
        # we'll actually get revision information in the file. If not,
        # we'll match self.BINARY_STRING.
        if line.startswith(b'Binary files') and line.endswith(b'differ'):
            linenum += 1

            parsed_file.binary = True
            parsed_file.orig_filename = file_index
            parsed_file.modified_filename = file_index

            m = _binary_revision_re.findall(line)

            if len(m) == 2:
                orig, modified = m

                parsed_file.orig_file_details = \
                    self.parse_diff_revision(file_index, orig[0])[1]
                parsed_file.modified_file_details = \
                    self.parse_diff_revision(file_index, modified[0])[1]
            else:
                parsed_file.orig_file_details = b'(unknown)'
                parsed_file.modified_file_details = b'(working copy)'
        elif lines[linenum] == self.BINARY_STRING:
            # Skip this and the svn:mime-type line.
            linenum += 2

            parsed_file.binary = True
            parsed_file.orig_filename = file_index
            parsed_file.modified_filename = file_index

            # We can't get the revision info from this diff header.
            parsed_file.orig_file_details = b'(unknown)'
            parsed_file.modified_file_details = b'(working copy)'

        return linenum

    def parse_after_headers(
        self,
        linenum: int,
        parsed_file: ParsedDiffFile,
    ) -> int:
        """Parse information after a diff header but before diff data.

        This looks for any property changes after the diff header, starting
        with ``Property changes on:``. If found, the changes will be kept
        only if processing a binary file, as binary files nearly always have
        a property that's relevant (the mimetype). Otherwise, property changes
        are skipped.

        Args:
            linenum (int):
                The line number to begin parsing.

            parsed_file (ParsedDiffFile):
                The file currently being parsed.

        Returns:
            int:
            The next line number to parse.
        """
        # We're looking for a SVN property change for SVN 1.7+.
        #
        # This differs from SVN property changes in older versions of SVN
        # in a couple ways:
        #
        # 1) The ---, +++, and Index: lines have actual filenames.
        #    Because of this, we won't hit the case in parse_diff_header
        #    above.
        # 2) There's an actual section per-property, so we could parse these
        #    out in a usable form. We'd still need a way to display that
        #    sanely, though.
        lines = self.lines

        try:
            if (lines[linenum] == b'' and
                lines[linenum + 1].startswith(b'Property changes on:')):
                # If we're working with binary files, we're going to leave
                # the data here and not skip the entry. SVN diffs may include
                # property changes as part of the binary file entry.
                if not parsed_file.binary:
                    # Skip over the next 3 lines (blank, "Property changes
                    # on:", and the "__________" divider.
                    parsed_file.skip = True

                linenum += 3
        except IndexError:
            pass

        return linenum


def recompute_svn_backend() -> None:
    """Recompute the SVNTool client backend to use.

    Normally, this is only called once, but it may be used to reset the
    backend for use in testing.
    """
    global _SVNClientBackend
    global has_svn_backend

    _SVNClientBackend = None
    has_svn_backend = False

    for backend_path in settings.SVNTOOL_BACKENDS:
        try:
            mod = import_module(backend_path)

            # Check that this is a valid SVN backend.
            if (not hasattr(mod, 'has_svn_backend') or
                not hasattr(mod, 'Client')):
                logger.error('Attempted to load invalid SVN backend %s',
                             backend_path)
                continue

            has_svn_backend = mod.has_svn_backend

            # We want either the winning SVN backend or the first one to show
            # up in the required module dependencies list.
            if has_svn_backend:
                SVNTool.dependencies['modules'] = [mod.Client.required_module]

            if has_svn_backend:
                # We found a suitable backend.
                logger.debug('Using %s backend for SVN', backend_path)
                _SVNClientBackend = mod.Client
                break
        except ImportError:
            logger.exception('Unable to load SVN backend %s', backend_path)


recompute_svn_backend()



"""A hook for setting access states on extra data fields."""

from __future__ import annotations

from djblets.extensions.hooks import ExtensionHook, ExtensionHookPoint
from djblets.registries.errors import ItemLookupError


class APIExtraDataAccessHook(ExtensionHook, metaclass=ExtensionHookPoint):
    """A hook for setting access states on extra data fields.

    Extensions can use this hook to register ``extra_data`` fields with
    certain access states on subclasses of
    :py:data:`~reviewboard.webapi.base.WebAPIResource`.

    This accepts a list of ``field_set`` values specified by the Extension and
    registers them when the hook is created. Likewise, it unregisters the same
    list of ``field_set`` values when the Extension is disabled.

    Each element of ``field_set`` is a 2-:py:class:`tuple` where the first
    element of the tuple is the field's path (as a :py:class:`tuple`) and the
    second is the field's access state (as one of
    :py:data:`~reviewboard.webapi.base.ExtraDataAccessLevel.ACCESS_STATE_PUBLIC`
    or :py:data:`~reviewboard.webapi.base.ExtraDataAccessLevel.ACCESS_STATE_PRIVATE`).

    Example:
        .. code-block:: python

            obj.extra_data = {
                'foo': {
                    'bar' : 'private_data',
                    'baz' : 'public_data'
                }
            }

            ...

            APIExtraDataAccessHook(
                extension,
                resource,
                [
                    (('foo', 'bar'), ExtraDataAccessLevel.ACCESS_STATE_PRIVATE,
                ])
    """

    def initialize(self, resource, field_set):
        """Initialize the APIExtraDataAccessHook.

        Args:
            resource (reviewboard.webapi.base.WebAPIResource):
                The resource to modify access states for.

            field_set (list):
                Each element of ``field_set`` is a 2-:py:class:`tuple` where
                the first element of the tuple is the field's path (as a
                :py:class:`tuple`) and the second is the field's access state
                (as one of
                :py:data:`~reviewboard.webapi.base.ExtraDataAccessLevel.ACCESS_STATE_PUBLIC`
                or :py:data:`~reviewboard.webapi.base.ExtraDataAccessLevel.ACCESS_STATE_PRIVATE`).
        """
        self.resource = resource
        self.field_set = field_set

        resource.extra_data_access_callbacks.register(
            self.get_extra_data_state)

    def get_extra_data_state(self, key_path):
        """Return the state of an extra_data field.

        Args:
            key_path (tuple):
                A tuple of strings representing the path of an extra_data
                field.

        Returns:
            int:
            The access state of the provided field or ``None``.
        """
        for path, access_state in self.field_set:
            if path == key_path:
                return access_state

        return None

    def shutdown(self):
        """Shut down the hook.

        This will unregister the access levels from the resource.
        """
        try:
            self.resource.extra_data_access_callbacks.unregister(
                self.get_extra_data_state)
        except ItemLookupError:
            pass



"""Hooks for working with user-related information.

Version Added:
    8.0
"""

from __future__ import annotations

from djblets.extensions.hooks import BaseRegistryHook, ExtensionHookPoint

from reviewboard.accounts.service_accounts import (ServiceAccount,
                                                   service_account_registry)
from reviewboard.accounts.user_details import (BaseUserDetailsProvider,
                                               user_details_provider_registry)


class UserDetailsProviderHook(BaseRegistryHook[BaseUserDetailsProvider],
                              metaclass=ExtensionHookPoint):
    """Hook to register custom user detail providers.

    See :ref:`user-details-provider-hook` for instructions.

    Version Added:
        8.0
    """

    registry = user_details_provider_registry


class ServiceAccountHook(BaseRegistryHook[ServiceAccount],
                         metaclass=ExtensionHookPoint):
    """Hook to register custom service accounts.

    See :ref:`service-account-hook` for instructions.

    Version Added:
        8.1
    """

    registry = service_account_registry



from __future__ import annotations

from typing import TYPE_CHECKING

from django_assert_queries.testing import assert_queries
from django.db.models import Q
from djblets.webapi.errors import INVALID_FORM_DATA, PERMISSION_DENIED

from reviewboard.attachments.models import (FileAttachment,
                                            FileAttachmentHistory)
from reviewboard.reviews.models import (
    ReviewRequest,
    ReviewRequestDraft)
from reviewboard.reviews.models.review_request import FileAttachmentState
from reviewboard.webapi.resources import resources
from reviewboard.webapi.testing.queries import (
    get_webapi_request_start_equeries,
)
from reviewboard.webapi.tests.base import BaseWebAPITestCase
from reviewboard.webapi.tests.mimetypes import (file_attachment_item_mimetype,
                                                file_attachment_list_mimetype)
from reviewboard.webapi.tests.mixins import (BasicTestsMetaclass,
                                             ReviewRequestChildItemMixin,
                                             ReviewRequestChildListMixin)
from reviewboard.webapi.tests.mixins_extra_data import (ExtraDataItemMixin,
                                                        ExtraDataListMixin)
from reviewboard.webapi.tests.urls import (get_file_attachment_item_url,
                                           get_file_attachment_list_url)

if TYPE_CHECKING:
    from django_assert_queries.query_comparator import ExpectedQueries


class ResourceListTests(ReviewRequestChildListMixin,
                        BaseWebAPITestCase,
                        ExtraDataListMixin,
                        metaclass=BasicTestsMetaclass):
    """Testing the FileAttachmentResource list APIs."""
    fixtures = ['test_users']
    basic_get_fixtures = ['test_scmtools']
    sample_api_url = 'review-requests/<id>/file-attachments/'
    resource = resources.file_attachment

    def setup_review_request_child_test(self, review_request):
        return (get_file_attachment_list_url(review_request),
                file_attachment_list_mimetype)

    def setup_http_not_allowed_item_test(self, user):
        review_request = self.create_review_request(
            submitter=user,
            publish=True)

        return get_file_attachment_list_url(review_request)

    def compare_item(self, item_rsp, attachment):
        self.assertEqual(item_rsp['id'], attachment.pk)
        self.assertEqual(item_rsp['is_review_ui_accessible_by'],
                         attachment.is_review_ui_accessible_by(self.user))
        self.assertEqual(
            item_rsp['extra_data'],
            self.resource.serialize_extra_data_field(attachment))
        self.assertEqual(item_rsp['filename'], attachment.filename)
        self.assertEqual(item_rsp['revision'], attachment.attachment_revision)
        self.assertEqual(item_rsp['sha256_checksum'],
                         attachment.sha256_checksum)

    #
    # HTTP GET tests
    #

    def setup_basic_get_test(self, user, with_local_site, local_site_name,
                             populate_items):
        review_request = self.create_review_request(
            create_repository=True,
            with_local_site=with_local_site,
            submitter=user)

        if populate_items:
            # This is the file attachment that should be returned.
            items = [
                self.create_file_attachment(review_request,
                                            orig_filename='logo1.png'),
            ]

            # This attachment shouldn't be shown in the results. It represents
            # a file to be shown in the diff viewer.
            self.create_file_attachment(review_request,
                                        orig_filename='logo2.png',
                                        repo_path='/logo.png',
                                        repo_revision='123',
                                        repository=review_request.repository)

            # This attachment shouldn't be shown either, for the same
            # reasons.
            diffset = self.create_diffset(review_request)
            filediff = self.create_filediff(diffset,
                                            source_file='/logo3.png',
                                            dest_file='/logo3.png',
                                            source_revision='123',
                                            dest_detail='124')
            self.create_file_attachment(review_request,
                                        orig_filename='logo3.png',
                                        added_in_filediff=filediff)
        else:
            items = []

        return (get_file_attachment_list_url(review_request, local_site_name),
                file_attachment_list_mimetype,
                items)

    #
    # HTTP POST tests
    #

    def setup_basic_post_test(self, user, with_local_site, local_site_name,
                              post_valid_data):
        review_request = self.create_review_request(
            with_local_site=with_local_site,
            submitter=user,
            publish=True)

        return (get_file_attachment_list_url(review_request, local_site_name),
                file_attachment_item_mimetype,
                {'path': open(self.get_sample_image_filename(), 'rb')},
                [review_request])

    def check_post_result(self, user, rsp, review_request):
        draft = review_request.get_draft()
        self.assertIsNotNone(draft)

        self.assertIn('file_attachment', rsp)
        item_rsp = rsp['file_attachment']

        attachment = FileAttachment.objects.get(pk=item_rsp['id'])
        self.assertIn(attachment, draft.file_attachments.all())
        self.assertNotIn(attachment, review_request.file_attachments.all())
        self.compare_item(item_rsp, attachment)

    def test_post_not_owner(self):
        """Testing the POST review-requests/<id>/file-attachments/ API
        without owner
        """
        review_request = self.create_review_request()
        self.assertNotEqual(review_request.submitter, self.user)

        with open(self.get_sample_image_filename(), 'rb') as f:
            self.assertTrue(f)
            rsp = self.api_post(
                get_file_attachment_list_url(review_request),
                {
                    'caption': 'logo',
                    'path': f,
                },
                expected_status=403)

        self.assertEqual(rsp['stat'], 'fail')
        self.assertEqual(rsp['err']['code'], PERMISSION_DENIED.code)

    def test_post_with_attachment_history_id(self):
        """Testing the POST review-requests/<id>/file-attachments/ API with a
        file attachment history
        """
        review_request = self.create_review_request(
            submitter=self.user, publish=True, target_people=[self.user])
        history = FileAttachmentHistory.objects.create(display_position=0)
        review_request.file_attachment_histories.add(history)

        self.assertEqual(history.latest_revision, 0)

        with open(self.get_sample_image_filename(), 'rb') as f:
            self.assertTrue(f)
            rsp = self.api_post(
                get_file_attachment_list_url(review_request),
                {
                    'path': f,
                    'attachment_history': history.pk,
                },
                expected_mimetype=file_attachment_item_mimetype)

            self.assertEqual(rsp['stat'], 'ok')
            self.assertEqual(rsp['file_attachment']['attachment_history_id'],
                             history.pk)

            history = FileAttachmentHistory.objects.get(pk=history.pk)
            self.assertEqual(history.latest_revision, 1)

            review_request.get_draft().publish()

            # Add a second revision
            f.seek(0)
            rsp = self.api_post(
                get_file_attachment_list_url(review_request),
                {
                    'path': f,
                    'attachment_history': history.pk,
                },
                expected_mimetype=file_attachment_item_mimetype)

            self.assertEqual(rsp['stat'], 'ok')
            self.assertEqual(rsp['file_attachment']['attachment_history_id'],
                             history.pk)

            history = FileAttachmentHistory.objects.get(pk=history.pk)
            self.assertEqual(history.latest_revision, 2)

    def test_post_with_attachment_history_id_wrong_review_request(self):
        """Testing the POST review-requests/<id>/file-attachments/ API with a
        file attachment history belonging to a different reiew request
        """
        review_request_1 = self.create_review_request(submitter=self.user,
                                                      publish=True)
        history = FileAttachmentHistory.objects.create(display_position=0)
        review_request_1.file_attachment_histories.add(history)

        review_request_2 = self.create_review_request(submitter=self.user,
                                                      publish=True)

        self.assertEqual(history.latest_revision, 0)

        with open(self.get_sample_image_filename(), 'rb') as f:
            self.assertTrue(f)
            rsp = self.api_post(
                get_file_attachment_list_url(review_request_2),
                {
                    'path': f,
                    'attachment_history': history.pk,
                },
                expected_status=400)

            self.assertEqual(rsp['stat'], 'fail')
            self.assertEqual(rsp['err']['code'], INVALID_FORM_DATA.code)

            history = FileAttachmentHistory.objects.get(pk=history.pk)
            self.assertEqual(history.latest_revision, 0)


class ResourceItemTests(ReviewRequestChildItemMixin,
                        BaseWebAPITestCase,
                        ExtraDataItemMixin,
                        metaclass=BasicTestsMetaclass):
    """Testing the FileAttachmentResource item APIs."""
    fixtures = ['test_users']
    sample_api_url = 'review-requests/<id>/file-attachments/<id>/'
    resource = resources.file_attachment

    default_extra_data = {
        FileAttachment.DEFINED_LOCAL_SITE_KEY: True,
        'sha256_checksum': ('1931a3b367e2913d28f9587dbd0ccf79b2c'
                            '2225de7c47550dd1cc49085077e49')
    }

    def setup_review_request_child_test(self, review_request):
        file_attachment = self.create_file_attachment(review_request)

        return (get_file_attachment_item_url(file_attachment),
                file_attachment_item_mimetype)

    def compare_item(self, item_rsp, attachment):
        self.assertEqual(item_rsp['id'], attachment.pk)
        self.assertEqual(item_rsp['is_review_ui_accessible_by'],
                         attachment.is_review_ui_accessible_by(self.user))
        self.assertEqual(
            item_rsp['extra_data'],
            self.resource.serialize_extra_data_field(attachment))
        self.assertEqual(item_rsp['filename'], attachment.filename)
        self.assertEqual(item_rsp['revision'], attachment.attachment_revision)
        self.assertEqual(item_rsp['absolute_url'],
                         attachment.get_absolute_url())

    def setup_http_not_allowed_list_test(self, user):
        review_request = self.create_review_request(
            submitter=user)
        file_attachment = self.create_file_attachment(review_request)

        return get_file_attachment_item_url(file_attachment)

    #
    # HTTP DELETE tests
    #

    def setup_basic_delete_test(self, user, with_local_site, local_site_name):
        review_request = self.create_review_request(
            with_local_site=with_local_site,
            submitter=user)
        file_attachment = self.create_file_attachment(review_request)

        return (get_file_attachment_item_url(file_attachment, local_site_name),
                [review_request, file_attachment])

    def check_delete_result(self, user, review_request, file_attachment):
        draft = review_request.get_draft()
        self.assertIsNotNone(draft)
        self.assertIn(file_attachment, draft.inactive_file_attachments.all())
        self.assertNotIn(file_attachment, draft.file_attachments.all())
        self.assertIn(file_attachment, review_request.file_attachments.all())

    #
    # HTTP GET tests
    #

    def setup_basic_get_test(self, user, with_local_site, local_site_name):
        review_request = self.create_review_request(
            with_local_site=with_local_site,
            submitter=user)
        file_attachment = self.create_file_attachment(review_request)

        return (get_file_attachment_item_url(file_attachment, local_site_name),
                file_attachment_item_mimetype,
                file_attachment)

    def test_get_not_modified(self):
        """Testing the GET review-requests/<id>/file-attachments/<id>/ API
        with Not Modified response
        """
        review_request = self.create_review_request(publish=True)
        file_attachment = self.create_file_attachment(review_request)

        self._testHttpCaching(get_file_attachment_item_url(file_attachment),
                              check_etags=True)

    #
    # HTTP PUT tests
    #

    def setup_basic_put_test(self, user, with_local_site, local_site_name,
                             put_valid_data):
        review_request = self.create_review_request(
            with_local_site=with_local_site,
            submitter=user)
        file_attachment = self.create_file_attachment(review_request)

        return (get_file_attachment_item_url(file_attachment, local_site_name),
                file_attachment_item_mimetype,
                {'caption': 'My new caption'},
                file_attachment,
                [review_request])

    def check_put_result(self, user, item_rsp, file_attachment,
                         review_request):
        file_attachment = FileAttachment.objects.get(pk=file_attachment.pk)
        self.assertEqual(item_rsp['id'], file_attachment.pk)
        self.assertEqual(file_attachment.draft_caption, 'My new caption')

        draft = review_request.get_draft()
        self.assertIsNotNone(draft)

        self.assertIn(file_attachment, draft.file_attachments.all())
        self.assertIn(file_attachment, review_request.file_attachments.all())
        self.compare_item(item_rsp, file_attachment)

    def test_put_with_no_sha256_checksum(self) -> None:
        """Testing the PUT review-requests/<id>/draft/file-attachments/<id>/
        for an attachment that doesn't have its sha256_checksum set already
        """
        user = self.user
        assert user is not None

        review_request = self.create_review_request(submitter=user)
        file_attachment = self.create_file_attachment(review_request)
        review_request_draft = self.create_review_request_draft(review_request)

        file_attachment.extra_data.pop('sha256_checksum')
        file_attachment.save(update_fields=('extra_data',))

        equeries: ExpectedQueries = [
            {
                'join_types': {
                    'reviews_reviewrequest_file_attachments': 'INNER JOIN',
                },
                'model': ReviewRequest,
                'num_joins': 1,
                'tables': {
                    'reviews_reviewrequest',
                    'reviews_reviewrequest_file_attachments',
                },
                'where': Q(file_attachments__id=file_attachment.pk),
            },
        ]
        equeries += get_webapi_request_start_equeries(user=user)
        equeries += [
            {
                'model': ReviewRequest,
                'select_related': {'submitter', 'repository'},
                'where': (Q(local_site=None) &
                          Q(pk=str(review_request.pk))),
            },
            {
                'model': ReviewRequestDraft,
                'select_related': {'review_request'},
                'where': Q(review_request=review_request),
            },
            {
                'join_types': {
                    'reviews_reviewrequest_file_attachments':
                        'LEFT OUTER JOIN',
                    'reviews_reviewrequestdraft_file_attachments':
                        'LEFT OUTER JOIN',
                },
                'model': FileAttachment,
                'num_joins': 2,
                'tables': {
                    'attachments_fileattachment',
                    'reviews_reviewrequest_file_attachments',
                    'reviews_reviewrequestdraft_file_attachments',
                },
                'where': (
                    Q(Q(Q(review_request=review_request) |
                        Q(drafts=review_request_draft)) &
                      Q(added_in_filediff__isnull=True) &
                      Q(repository__isnull=True) &
                      Q(user__isnull=True)) &
                    Q(pk=str(review_request.pk))
                ),
            },
            {
                'extra': {
                    '_prefetch_related_val_fileattachment_id': (
                        '"reviews_reviewrequest_file_attachments".'
                        '"fileattachment_id"',
                        [],
                    ),
                },
                'join_types': {
                    'reviews_reviewrequest_file_attachments': 'INNER JOIN',
                },
                'model': ReviewRequest,
                'num_joins': 1,
                'tables': {
                    'reviews_reviewrequest',
                    'reviews_reviewrequest_file_attachments',
                },
                'where': Q(file_attachments__in=[file_attachment]),
            },
            {
                'extra': {
                    '_prefetch_related_val_fileattachment_id': (
                        '"reviews_reviewrequestdraft_file_attachments".'
                        '"fileattachment_id"',
                        [],
                    ),
                },
                'join_types': {
                    'reviews_reviewrequestdraft_file_attachments':
                        'INNER JOIN',
                },
                'model': ReviewRequestDraft,
                'num_joins': 1,
                'tables': {
                    'reviews_reviewrequestdraft',
                    'reviews_reviewrequestdraft_file_attachments',
                },
                'where': Q(file_attachments__in=[file_attachment]),
            },
            {
                'model': ReviewRequestDraft,
                'where': Q(review_request=review_request),
            },
            {
                'model': FileAttachment,
                'type': 'UPDATE',
                'where': Q(pk=file_attachment.pk),
            },
            {
                'model': ReviewRequestDraft,
                'type': 'UPDATE',
                'where': Q(pk=review_request_draft.pk)
            },

            # This extra query is for setting sha256_checksum in extra_data,
            # which happens when we serialize the file attachment.
            {
                'model': FileAttachment,
                'type': 'UPDATE',
                'where': Q(pk=file_attachment.pk),
            },
        ]

        with assert_queries(equeries):
            rsp = self.api_put(
                get_file_attachment_item_url(file_attachment),
                {
                    'caption': 'Updated caption',
                },
                expected_mimetype=file_attachment_item_mimetype)

        self.assertEqual(rsp['stat'], 'ok')

    def test_put_with_caption(self) -> None:
        """Testing the PUT review-requests/<id>/draft/file-attachments/<id>/
        with setting a new caption and when a draft already exists
        """
        user = self.user
        assert user is not None

        review_request = self.create_review_request(submitter=user)
        file_attachment = self.create_file_attachment(review_request)
        review_request_draft = self.create_review_request_draft(review_request)

        equeries: ExpectedQueries = [
            {
                'join_types': {
                    'reviews_reviewrequest_file_attachments': 'INNER JOIN',
                },
                'model': ReviewRequest,
                'num_joins': 1,
                'tables': {
                    'reviews_reviewrequest',
                    'reviews_reviewrequest_file_attachments',
                },
                'where': Q(file_attachments__id=file_attachment.pk),
            },
        ]
        equeries += get_webapi_request_start_equeries(user=user)
        equeries += [
            {
                'model': ReviewRequest,
                'select_related': {'submitter', 'repository'},
                'where': (Q(local_site=None) &
                          Q(pk=str(review_request.pk))),
            },
            {
                'model': ReviewRequestDraft,
                'select_related': {'review_request'},
                'where': Q(review_request=review_request),
            },
            {
                'join_types': {
                    'reviews_reviewrequest_file_attachments':
                        'LEFT OUTER JOIN',
                    'reviews_reviewrequestdraft_file_attachments':
                        'LEFT OUTER JOIN',
                },
                'model': FileAttachment,
                'num_joins': 2,
                'tables': {
                    'attachments_fileattachment',
                    'reviews_reviewrequest_file_attachments',
                    'reviews_reviewrequestdraft_file_attachments',
                },
                'where': (
                    Q(Q(Q(review_request=review_request) |
                        Q(drafts=review_request_draft)) &
                      Q(added_in_filediff__isnull=True) &
                      Q(repository__isnull=True) &
                      Q(user__isnull=True)) &
                    Q(pk=str(review_request.pk))
                ),
            },
            {
                'extra': {
                    '_prefetch_related_val_fileattachment_id': (
                        '"reviews_reviewrequest_file_attachments".'
                        '"fileattachment_id"',
                        [],
                    ),
                },
                'join_types': {
                    'reviews_reviewrequest_file_attachments': 'INNER JOIN',
                },
                'model': ReviewRequest,
                'num_joins': 1,
                'tables': {
                    'reviews_reviewrequest',
                    'reviews_reviewrequest_file_attachments',
                },
                'where': Q(file_attachments__in=[file_attachment]),
            },
            {
                'extra': {
                    '_prefetch_related_val_fileattachment_id': (
                        '"reviews_reviewrequestdraft_file_attachments".'
                        '"fileattachment_id"',
                        [],
                    ),
                },
                'join_types': {
                    'reviews_reviewrequestdraft_file_attachments':
                        'INNER JOIN',
                },
                'model': ReviewRequestDraft,
                'num_joins': 1,
                'tables': {
                    'reviews_reviewrequestdraft',
                    'reviews_reviewrequestdraft_file_attachments',
                },
                'where': Q(file_attachments__in=[file_attachment]),
            },
            {
                'model': ReviewRequestDraft,
                'where': Q(review_request=review_request),
            },
            {
                'model': FileAttachment,
                'type': 'UPDATE',
                'where': Q(pk=file_attachment.pk),
            },
            {
                'model': ReviewRequestDraft,
                'type': 'UPDATE',
                'where': Q(pk=review_request_draft.pk)
            },
        ]

        # The purpose of this test is to see what queries are being executed,
        # to compare against the queries executed during ``pending_deletion``
        # updates.
        with assert_queries(equeries):
            rsp = self.api_put(
                get_file_attachment_item_url(file_attachment),
                {
                    'caption': 'Updated caption',
                },
                expected_mimetype=file_attachment_item_mimetype)

        self.assertEqual(rsp['stat'], 'ok')

    def test_put_with_pending_deletion_false(self) -> None:
        """Testing the PUT review-requests/<id>/draft/file-attachments/<id>/
        with setting pending_deletion to False for a file attachment that
        is currently pending deletion
        """
        user = self.user
        assert user is not None

        review_request = self.create_review_request(submitter=user)
        file_attachment = self.create_file_attachment(review_request)
        review_request_draft = self.create_review_request_draft(review_request)

        # "Delete" the file attachment.
        review_request_draft.inactive_file_attachments.add(file_attachment)
        review_request_draft.file_attachments.remove(file_attachment)

        self.assertEqual(
            review_request.get_file_attachment_state(file_attachment),
            FileAttachmentState.PENDING_DELETION)

        del review_request._file_attachments_data

        equeries: ExpectedQueries = [
            {
                'join_types': {
                    'reviews_reviewrequest_file_attachments': 'INNER JOIN',
                },
                'model': ReviewRequest,
                'num_joins': 1,
                'tables': {
                    'reviews_reviewrequest',
                    'reviews_reviewrequest_file_attachments',
                },
                'where': Q(file_attachments__id=file_attachment.pk),
            },
        ]
        equeries += get_webapi_request_start_equeries(user=user)
        equeries += [
            {
                'model': ReviewRequest,
                'select_related': {'submitter', 'repository'},
                'where': (Q(local_site=None) &
                          Q(pk=str(review_request.pk))),
            },
            {
                'model': ReviewRequestDraft,
                'select_related': {'review_request'},
                'where': Q(review_request=review_request),
            },
            {
                'join_types': {
                    'reviews_reviewrequest_file_attachments':
                        'LEFT OUTER JOIN',
                    'reviews_reviewrequestdraft_inactive_file_attachments':
                        'LEFT OUTER JOIN',
                },
                'model': FileAttachment,
                'num_joins': 2,
                'tables': {
                    'attachments_fileattachment',
                    'reviews_reviewrequest_file_attachments',
                    'reviews_reviewrequestdraft_inactive_file_attachments',
                },
                'where': (
                    Q(Q(Q(review_request=review_request) |
                        Q(inactive_drafts=review_request_draft)) &
                      Q(added_in_filediff__isnull=True) &
                      Q(repository__isnull=True) &
                      Q(user__isnull=True)) &
                    Q(pk=str(review_request.pk))
                ),
            },
            {
                'extra': {
                    '_prefetch_related_val_fileattachment_id': (
                        '"reviews_reviewrequest_file_attachments".'
                        '"fileattachment_id"',
                        [],
                    ),
                },
                'join_types': {
                    'reviews_reviewrequest_file_attachments': 'INNER JOIN',
                },
                'model': ReviewRequest,
                'num_joins': 1,
                'tables': {
                    'reviews_reviewrequest',
                    'reviews_reviewrequest_file_attachments',
                },
                'where': Q(file_attachments__in=[file_attachment]),
            },
            {
                'extra': {
                    '_prefetch_related_val_fileattachment_id': (
                        '"reviews_reviewrequestdraft_inactive_file_'
                        'attachments"."fileattachment_id"',
                        [],
                    ),
                },
                'join_types': {
                    'reviews_reviewrequestdraft_inactive_file_attachments':
                        'INNER JOIN',
                },
                'model': ReviewRequestDraft,
                'num_joins': 1,
                'tables': {
                    'reviews_reviewrequestdraft',
                    'reviews_reviewrequestdraft_inactive_file_attachments',
                },
                'where': Q(inactive_file_attachments__in=[file_attachment]),
            },
            {
                'model': ReviewRequestDraft,
                'where': Q(review_request=review_request),
            },
            {
                'model': ReviewRequest,
                'where': Q(id=review_request.pk),
            },
            {
                'join_types': {
                    'reviews_reviewrequest_file_attachments': 'INNER JOIN',
                },
                'model': FileAttachment,
                'num_joins': 1,
                'tables': {
                    'attachments_fileattachment',
                    'reviews_reviewrequest_file_attachments',
                },
                'where': Q(review_request__id=review_request.pk),
            },
            {
                'model': ReviewRequestDraft,
                'where': Q(review_request=review_request),
            },
            {
                'join_types': {
                    'reviews_reviewrequestdraft_inactive_file_attachments':
                        'INNER JOIN',
                },
                'model': FileAttachment,
                'num_joins': 1,
                'tables': {
                    'attachments_fileattachment',
                    'reviews_reviewrequestdraft_inactive_file_attachments',
                },
                'where': Q(inactive_drafts__id=review_request_draft.pk),
            },
            {
                'model': FileAttachment,
                'where': Q(
                    attachment_history=file_attachment.attachment_history_id),
                'values_select': ('pk',),
            },
            {
                'model': ReviewRequestDraft.inactive_file_attachments.through,
                'where': (Q(reviewrequestdraft=(review_request_draft.pk,)) &
                          Q(fileattachment__in={file_attachment.pk})),
            },
            {
                'model': ReviewRequestDraft.inactive_file_attachments.through,
                'type': 'DELETE',
                'where': Q(id__in=[file_attachment.pk]),
            },
            {
                'model': ReviewRequestDraft,
                'type': 'UPDATE',
                'where': Q(pk=review_request_draft.pk),
            },
            {
                'model': ReviewRequestDraft,
                'limit': 1,
                'values_select': ('inactive_file_attachments_count',),
                'where': Q(pk=review_request_draft.pk),
            },
            {
                'model': ReviewRequestDraft.file_attachments.through,
                'values_select': ('fileattachment',),
                'where': (Q(fileattachment__in={file_attachment.pk}) &
                          Q(reviewrequestdraft=review_request_draft.pk)),
            },
            {
                'model': ReviewRequestDraft.file_attachments.through,
                'type': 'INSERT',
            },
            {
                'model': ReviewRequestDraft,
                'type': 'UPDATE',
                'where': Q(pk=review_request_draft.pk),
            },
            {
                'model': ReviewRequestDraft,
                'limit': 1,
                'values_select': ('file_attachments_count',),
                'where': Q(pk=review_request_draft.pk),
            },
            {
                'model': ReviewRequestDraft,
                'type': 'UPDATE',
                'where': Q(pk=review_request_draft.pk)
            },
        ]

        with assert_queries(equeries):
            rsp = self.api_put(
                get_file_attachment_item_url(file_attachment),
                {
                    'pending_deletion': False,
                },
                expected_mimetype=file_attachment_item_mimetype)

        self.assertEqual(rsp['stat'], 'ok')
        self.assertEqual(
            review_request.get_file_attachment_state(file_attachment),
            FileAttachmentState.PUBLISHED)

    def test_put_with_pending_deletion_false_with_history(self) -> None:
        """Testing the PUT review-requests/<id>/draft/file-attachments/<id>/
        with setting pending_deletion to False for a file attachment that
        is currently pending deletion and has a history of revisions
        """
        user = self.user
        assert user is not None

        review_request = self.create_review_request(submitter=user)
        file_attachment = self.create_file_attachment(review_request)
        file_attachment_2 = self.create_file_attachment(
            review_request,
            attachment_history=file_attachment.attachment_history,
            attachment_revision=file_attachment.attachment_revision + 1)
        review_request_draft = self.create_review_request_draft(review_request)

        # "Delete" the file attachment and all revisions of it.
        review_request_draft.inactive_file_attachments.add(file_attachment)
        review_request_draft.file_attachments.remove(file_attachment)
        review_request_draft.inactive_file_attachments.add(file_attachment_2)
        review_request_draft.file_attachments.remove(file_attachment_2)

        self.assertEqual(
            review_request.get_file_attachment_state(file_attachment),
            FileAttachmentState.PENDING_DELETION)
        self.assertEqual(
            review_request.get_file_attachment_state(file_attachment_2),
            FileAttachmentState.PENDING_DELETION)

        del review_request._file_attachments_data

        equeries: ExpectedQueries = [
            {
                'join_types': {
                    'reviews_reviewrequest_file_attachments': 'INNER JOIN',
                },
                'model': ReviewRequest,
                'num_joins': 1,
                'tables': {
                    'reviews_reviewrequest',
                    'reviews_reviewrequest_file_attachments',
                },
                'where': Q(file_attachments__id=file_attachment.pk),
            },
        ]
        equeries += get_webapi_request_start_equeries(user=user)
        equeries += [
            {
                'model': ReviewRequest,
                'select_related': {'submitter', 'repository'},
                'where': (Q(local_site=None) &
                          Q(pk=str(review_request.pk))),
            },
            {
                'model': ReviewRequestDraft,
                'select_related': {'review_request'},
                'where': Q(review_request=review_request),
            },
            {
                'join_types': {
                    'reviews_reviewrequest_file_attachments':
                        'LEFT OUTER JOIN',
                    'reviews_reviewrequestdraft_inactive_file_attachments':
                        'LEFT OUTER JOIN',
                },
                'model': FileAttachment,
                'num_joins': 2,
                'tables': {
                    'attachments_fileattachment',
                    'reviews_reviewrequest_file_attachments',
                    'reviews_reviewrequestdraft_inactive_file_attachments',
                },
                'where': (
                    Q(Q(Q(review_request=review_request) |
                        Q(inactive_drafts=review_request_draft)) &
                      Q(added_in_filediff__isnull=True) &
                      Q(repository__isnull=True) &
                      Q(user__isnull=True)) &
                    Q(pk=str(review_request.pk))
                ),
            },
            {
                'extra': {
                    '_prefetch_related_val_fileattachment_id': (
                        '"reviews_reviewrequest_file_attachments".'
                        '"fileattachment_id"',
                        [],
                    ),
                },
                'join_types': {
                    'reviews_reviewrequest_file_attachments': 'INNER JOIN',
                },
                'model': ReviewRequest,
                'num_joins': 1,
                'tables': {
                    'reviews_reviewrequest',
                    'reviews_reviewrequest_file_attachments',
                },
                'where': Q(file_attachments__in=[file_attachment]),
            },
            {
                'extra': {
                    '_prefetch_related_val_fileattachment_id': (
                        '"reviews_reviewrequestdraft_inactive_file_'
                        'attachments"."fileattachment_id"',
                        [],
                    ),
                },
                'join_types': {
                    'reviews_reviewrequestdraft_inactive_file_attachments':
                        'INNER JOIN',
                },
                'model': ReviewRequestDraft,
                'num_joins': 1,
                'tables': {
                    'reviews_reviewrequestdraft',
                    'reviews_reviewrequestdraft_inactive_file_attachments',
                },
                'where': Q(inactive_file_attachments__in=[file_attachment]),
            },
            {
                'model': ReviewRequestDraft,
                'where': Q(review_request=review_request),
            },
            {
                'model': ReviewRequest,
                'where': Q(id=review_request.pk),
            },
            {
                'join_types': {
                    'reviews_reviewrequest_file_attachments': 'INNER JOIN',
                },
                'model': FileAttachment,
                'num_joins': 1,
                'tables': {
                    'attachments_fileattachment',
                    'reviews_reviewrequest_file_attachments',
                },
                'where': Q(review_request__id=review_request.pk),
            },
            {
                'model': ReviewRequestDraft,
                'where': Q(review_request=review_request),
            },
            {
                'join_types': {
                    'reviews_reviewrequestdraft_inactive_file_attachments':
                        'INNER JOIN',
                },
                'model': FileAttachment,
                'num_joins': 1,
                'tables': {
                    'attachments_fileattachment',
                    'reviews_reviewrequestdraft_inactive_file_attachments',
                },
                'where': Q(inactive_drafts__id=review_request_draft.pk),
            },
            {
                'model': FileAttachment,
                'where': Q(
                    attachment_history=file_attachment.attachment_history_id),
                'values_select': ('pk',),
            },
            {
                'model': ReviewRequestDraft.inactive_file_attachments.through,
                'where': (Q(reviewrequestdraft=(review_request_draft.pk,)) &
                          Q(fileattachment__in={file_attachment.pk,
                                                file_attachment_2.pk})),
            },
            {
                'model': ReviewRequestDraft.inactive_file_attachments.through,
                'type': 'DELETE',
                'where': Q(id__in=[file_attachment_2.pk, file_attachment.pk]),
            },
            {
                'model': ReviewRequestDraft,
                'type': 'UPDATE',
                'where': Q(pk=review_request_draft.pk),
            },
            {
                'model': ReviewRequestDraft,
                'limit': 1,
                'values_select': ('inactive_file_attachments_count',),
                'where': Q(pk=review_request_draft.pk),
            },
            {
                'model': ReviewRequestDraft.file_attachments.through,
                'values_select': ('fileattachment',),
                'where': (Q(fileattachment__in={file_attachment.pk,
                                                file_attachment_2.pk}) &
                          Q(reviewrequestdraft=review_request_draft.pk)),
            },
            {
                'model': ReviewRequestDraft.file_attachments.through,
                'type': 'INSERT',
            },
            {
                'model': ReviewRequestDraft,
                'type': 'UPDATE',
                'where': Q(pk=review_request_draft.pk),
            },
            {
                'model': ReviewRequestDraft,
                'limit': 1,
                'values_select': ('file_attachments_count',),
                'where': Q(pk=review_request_draft.pk),
            },
            {
                'model': ReviewRequestDraft,
                'type': 'UPDATE',
                'where': Q(pk=review_request_draft.pk)
            },
        ]

        with assert_queries(equeries):
            rsp = self.api_put(
                get_file_attachment_item_url(file_attachment),
                {
                    'pending_deletion': False,
                },
                expected_mimetype=file_attachment_item_mimetype)

        self.assertEqual(rsp['stat'], 'ok')
        self.assertEqual(
            review_request.get_file_attachment_state(file_attachment),
            FileAttachmentState.PUBLISHED)
        self.assertEqual(
            review_request.get_file_attachment_state(file_attachment_2),
            FileAttachmentState.PUBLISHED)

    def test_put_with_pending_deletion_false_for_invalid(self) -> None:
        """Testing the PUT review-requests/<id>/draft/file-attachments/<id>/
        with setting pending_deletion to False for a file attachment that isn't
        currently pending deletion
        """
        review_request = self.create_review_request(submitter=self.user)
        file_attachment = self.create_file_attachment(review_request)
        self.create_review_request_draft(review_request)

        rsp = self.api_put(
            get_file_attachment_item_url(file_attachment),
            {
                'pending_deletion': False,
            },
            expected_status=400)

        self.assertEqual(
            rsp,
            {
                'err': {
                    'code': 105,
                    'msg': 'One or more fields had errors',
                    'type': 'request-field-error',
                },
                'fields': {
                    'pending_deletion': (
                        'This can only be used to undo the pending '
                        'deletion of a file attachment. This file '
                        'attachment is not currently pending '
                        'deletion.'
                    ),
                },
                'stat': 'fail',
            })



"""RawFileDiffData model definition."""

from __future__ import annotations

import bz2
import logging
from typing import ClassVar

from django.db import models
from django.utils.translation import gettext_lazy as _
from djblets.db.fields import JSONField

from reviewboard.diffviewer.errors import DiffParserError
from reviewboard.diffviewer.managers import RawFileDiffDataManager


logger = logging.getLogger(__name__)


class RawFileDiffData(models.Model):
    """Stores raw diff data as binary content in the database.

    This is the class used in Review Board 2.5+ to store diff content.
    Unlike in previous versions, the content is not base64-encoded. Instead,
    it is stored either as bzip2-compressed data (if the resulting
    compressed data is smaller than the raw data), or as the raw data itself.
    """

    COMPRESSION_BZIP2 = 'B'

    COMPRESSION_CHOICES = (
        (COMPRESSION_BZIP2, _('BZip2-compressed')),
    )

    binary_hash = models.CharField(_("hash"), max_length=40, unique=True)
    binary = models.BinaryField()
    compression = models.CharField(max_length=1, choices=COMPRESSION_CHOICES,
                                   null=True, blank=True)
    extra_data = JSONField(null=True)

    objects: ClassVar[RawFileDiffDataManager] = RawFileDiffDataManager()

    @property
    def content(self):
        """Return the content of the diff.

        The content will be uncompressed (if necessary) and returned as the
        raw set of bytes originally uploaded.
        """
        if self.compression == self.COMPRESSION_BZIP2:
            return bz2.decompress(self.binary)
        elif self.compression is None:
            return bytes(self.binary)
        else:
            raise NotImplementedError(
                'Unsupported compression method %s for RawFileDiffData %s'
                % (self.compression, self.pk))

    @property
    def insert_count(self):
        return self.extra_data.get('insert_count')

    @insert_count.setter
    def insert_count(self, value):
        self.extra_data['insert_count'] = value

    @property
    def delete_count(self):
        return self.extra_data.get('delete_count')

    @delete_count.setter
    def delete_count(self, value):
        self.extra_data['delete_count'] = value

    def recalculate_line_counts(self, tool):
        """Recalculates the insert_count and delete_count values.

        This will attempt to re-parse the stored diff and fetch the
        line counts through the parser.
        """
        logger.debug('Recalculating insert/delete line counts on '
                     'RawFileDiffData %s',
                     self.pk)

        try:
            files = tool.get_parser(self.content).parse()

            if len(files) != 1:
                raise DiffParserError(
                    'Got wrong number of files (%d)' % len(files))
        except DiffParserError as e:
            logger.error('Failed to correctly parse stored diff data in '
                         'RawFileDiffData ID %s when trying to get '
                         'insert/delete line counts: %s',
                         self.pk, e)
        else:
            file_info = files[0]
            self.insert_count = file_info.insert_count
            self.delete_count = file_info.delete_count

            if self.pk:
                self.save(update_fields=['extra_data'])

    class Meta:
        app_label = 'diffviewer'
        db_table = 'diffviewer_rawfilediffdata'
        verbose_name = _('Raw File Diff Data')
        verbose_name_plural = _('Raw File Diff Data Blobs')



"""Haystack engine that forwards to dynamically-configured engine."""

import logging
import threading

from djblets.siteconfig.models import SiteConfiguration
from haystack.backends import BaseEngine

from reviewboard.search import search_backend_registry


logger = logging.getLogger(__name__)


class BaseForwardingObject(object):
    """Base class for an object that forwards on to a search backend's object.

    This works as a proxy to an object on a configured search backend. The
    appropriate object will be constructed as needed and used for all method
    calls and attribute usage.

    Version Added:
        4.0
    """

    #: The name of the attribute on the engine class referencing the object.
    engine_attr = None

    def __init__(self, forwarding_engine, forwarded_obj=None,
                 forwarded_args=[], forwarded_kwargs={}):
        """Initialize the forwarding object.

        Args:
            forwarding_engine (ForwardingSearchEngine):
                The parent forwarding engine.

            forwarded_obj (object, optional):
                A specific instance of an object to forward to, if one is
                available.

            forwarded_args (tuple, optional):
                Positional arguments used to construct the forwarded object.

            forwarded_kwargs (dict, optional):
                Keyword arguments used to construct the forwarded object.
        """
        self._forwarding_engine = forwarding_engine
        self._forwarded_obj = forwarded_obj
        self._forwarded_args = forwarded_args
        self._forwarded_kwargs = forwarded_kwargs

    @property
    def forwarded_obj(self):
        """The object to forward all access to.

        This will be constructed and cached if not already set.

        Type:
            object
        """
        if self._forwarded_obj is None:
            forwarded_cls = getattr(self._forwarding_engine.forwarded_engine,
                                    self.engine_attr)

            self._forwarded_obj = forwarded_cls(*self._forwarded_args,
                                                **self._forwarded_kwargs)

        return self._forwarded_obj

    def __setattribute__(self, name, value):
        """Set an attribute.

        If the attribute belongs to this class, it will be set directly.
        Otherwise, it will be set on the forwarded object.

        Args:
            name (str):
                The name of the attribute.

            value (object):
                The value to set.
        """
        if name in self.__dict__:
            object.__setattribute__(self, name, value)
        else:
            setattr(self.forwarded_obj, name, value)

    def __getattribute__(self, name):
        """Return an attribute value.

        If the attribute belongs to this class, it will be returned directly.
        Otherwise, it will be returned from the forwarded object.

        Args:
            name (str):
                The name of the attribute.

        Returns:
            object:
            The attribute value.

        Raises:
            AttributeError:
                The attribute could not be found.
        """
        try:
            return object.__getattribute__(self, name)
        except AttributeError:
            return getattr(self.forwarded_obj, name)

    def __repr__(self):
        """Return a string representation of this object.

        Returns:
            unicode:
            The string representation.
        """
        return '<%s(%r)>' % (type(self).__name__, self.forwarded_obj)


class ForwardingSearchBackend(BaseForwardingObject):
    """A forwarding object for a Haystack search backend.

    This wraps a :py:class:`haystack.backends.BaseSearchBackend`.

    Version Added:
        4.0
    """

    engine_attr = 'backend'


class ForwardingSearchQuery(BaseForwardingObject):
    """A forwarding object for a Haystack search query.

    This wraps a :py:class:`haystack.backends.BaseSearchQuery`.

    Version Added:
        4.0
    """

    engine_attr = 'query'

    def _clone(self, klass=None, using=None):
        """Clone the object.

        This is an implementation of a private method in
        :py:class:`haystack.backends.BaseSearchQuery` to handle cloning of
        a query. It will return the clone wrapped in a forwarding object,
        or return a new forwarding object set up to construct an object if one
        doesn't already exist.

        Args:
            klass (type, optional):
                The desired class to construct for the forwarded query object.

            using (unicode, optional):
                The database backend to use. This is expected to be ``None``
                or "default".

        Returns:
            ForwardingSearchQuery:
            The cloned query object.
        """
        if self._forwarded_obj is not None:
            forwarded_obj = self._forwarded_obj._clone(klass=klass,
                                                       using=using)
        else:
            forwarded_obj = None

        return ForwardingSearchQuery(
            forwarding_engine=self._forwarding_engine,
            forwarded_obj=forwarded_obj,
            forwarded_args=self._forwarded_args,
            forwarded_kwargs=self._forwarded_kwargs)


class ForwardingUnifiedIndex(BaseForwardingObject):
    """A forwarding object for a Haystack search query.

    This wraps a :py:class:`haystack.utils.loading.UnifiedIndex`.

    Version Added:
        4.0
    """

    engine_attr = 'unified_index'


class ForwardingSearchEngine(BaseEngine):
    """A Haystack search engine that forwards to another engine.

    This is set as the default search engine for Haystack, and forwards all
    requests to a dynamically-configured search engine backend.

    A forwarding setup is necessary for dynamic configuration with Haystack,
    as it normally computes all backend state on import, and assumes it will
    never change.

    Along with this, some classes (like
    :py:class:`haystack.generic_views.SearchMixin`) set class-level instances
    that store permanent backend state. These must also be overridden to set
    this state per-request in order to gain the full benefits of this engine.

    Version Added:
        4.0
    """

    backend = ForwardingSearchBackend
    query = ForwardingSearchQuery
    unified_index = ForwardingUnifiedIndex

    def __init__(self, *args, **kwargs):
        """Initialize the forwarding engine.

        Args:
            *args (tuple):
                Positional arguments to pass to the parent class.

            **kwargs (dict):
                Keyword arguments to pass to the parent class.
        """
        super(ForwardingSearchEngine, self).__init__(*args, **kwargs)

        self._forwarded_engine = None
        self._load_lock = threading.Lock()
        self._load_gen = 0
        self._backend = None
        self._index = None
        self._engine_options = None

    @property
    def forwarded_engine(self):
        """The Haystack search engine requests are forwarded to.

        Type:
            haystack.backends.BaseEngine
        """
        if self._forwarded_engine is None:
            self._load_forwarded_engine()

        return self._forwarded_engine

    @property
    def forwarded_options(self):
        """The options for the forwarded Haystack search engine.

        Type:
            dict
        """
        if self._forwarded_engine is None:
            self._load_forwarded_engine()

        assert self._forwarded_options is not None
        return self._forwarded_options

    def get_backend(self):
        """Return the Haystack search backend.

        This will construct a backend object that forwards on to the configured
        Haystack search engine's version.

        Results are cached until the search configuration changes.

        Returns:
            ForwardingSearchBackend:
            The forwarding search backend.
        """
        if self._backend is None:
            self._backend = self.backend(
                forwarding_engine=self,
                forwarded_args=(self.using,),
                forwarded_kwargs=self.forwarded_options)

        return self._backend

    def get_query(self):
        """Return a Haystack search query.

        This will construct a query object that forwards on to the configured
        Haystack search engine's version.

        Results are not cached. This will construct a new instance each time,
        as per Haystack's default behavior.

        Returns:
            ForwardingSearchQuery:
            The forwarding search query.
        """
        return self.query(
            forwarding_engine=self,
            forwarded_kwargs={
                'using': self.using,
            })

    def get_unified_index(self):
        """Return the Haystack unified index object.

        This will construct a unified index object that forwards on to the
        configured Haystack search engine's version.

        Results are cached until the search configuration changes.

        Returns:
            ForwardingUnifiedIndex:
            The forwarding unified index.
        """
        if self._index is None:
            self._index = self.unified_index(
                forwarding_engine=self,
                forwarded_args=(
                    self.forwarded_options.get('EXCLUDED_INDEXES'),
                ))

        return self._index

    def reset_forwarding(self):
        """Reset the forwarding state.

        This should be used whenever we're applying a new search configuration.
        It will clear out any search sessions, the forwarding engine, and any
        cached objects, so that new state can be set on next use.
        """
        self._backend = None
        self._forwarded_engine = None
        self._forwarded_options = None
        self._index = None

        self.reset_sessions()

    def _load_forwarded_engine(self):
        """Load a forwarded Haystack search engine.

        This will look up the current site configuration and determine the
        proper Review Board and Haystack search backends to load, setting any
        forwarding state.

        IF there's an issue at all with loading the search backends, an error
        will be logged and search will be disabled.

        This is thread-safe.
        """
        cur_load_gen = self._load_gen

        with self._load_lock:
            if self._load_gen == cur_load_gen:
                self.reset_forwarding()

                siteconfig = SiteConfiguration.objects.get_current()
                search_backend_id = siteconfig.get('search_backend_id')
                search_backend = search_backend_registry.get_search_backend(
                    search_backend_id)

                if search_backend is None:
                    logger.error('The search engine "%s" could not be found. '
                                 'If this is provided by an extension, you '
                                 'will have to make sure that extension is '
                                 'enabled. Disabling search.',
                                 search_backend_id)
                    engine = None
                else:
                    try:
                        engine = search_backend.load_haystack_engine()
                    except Exception as e:
                        logger.error('Error loading the search engine "%s": '
                                     '%s',
                                     search_backend_id, e)
                        engine = None

                if engine is None:
                    # Disable search, since it's useless at this point.
                    siteconfig.set('search_enable', False)
                    siteconfig.save(update_fields=('settings',))
                else:
                    self._forwarded_engine = engine
                    self._search_backend = None
                    self._forwarded_options = search_backend.configuration

                self._load_gen = cur_load_gen + 1



"""Add FileDiffData.delete_count and insert_count.

Version Added:
    2.0
"""

from django_evolution.mutations import AddField
from django.db import models


MUTATIONS = [
    AddField('FileDiffData', 'insert_count', models.IntegerField, null=True),
    AddField('FileDiffData', 'delete_count', models.IntegerField, null=True)
]



"""Unit tests for reviewboard.codesafety.checkers.base.BaseCodeSafetyChecker.
"""

import kgb
from django.template.loader import render_to_string
from django.utils.translation import gettext_lazy

from reviewboard.codesafety.checkers.base import BaseCodeSafetyChecker
from reviewboard.testing import TestCase


class MyCodeSafetyChecker(BaseCodeSafetyChecker):
    checker_id = 'my-checker'
    summary = 'This is a test'
    result_labels = {
        'error1': gettext_lazy('Error #1'),
        'warning1': 'Warning #1',
    }


class BaseCodeSafetyCheckerTests(kgb.SpyAgency, TestCase):
    """Unit tests for BaseCodeSafetyChecker."""

    def setUp(self):
        super(BaseCodeSafetyCheckerTests, self).setUp()

        self.checker = MyCodeSafetyChecker()

    def test_get_result_labels(self):
        """Testing BaseCodeSafetyChecker.get_result_labels"""

        self.assertEqual(
            self.checker.get_result_labels(['warning1', 'error1', 'unknown']),
            ['Warning #1', 'Error #1', 'unknown'])

    def test_render_file_alert_html(self):
        """Testing BaseCodeSafetyChecker.render_file_alert_html"""
        class TemplateMyCodeSafetyChecker(MyCodeSafetyChecker):
            file_alert_html_template_name = 'my-alert.html'

        self.spy_on(render_to_string,
                    op=kgb.SpyOpReturn('<html>result</html>'))

        checker = TemplateMyCodeSafetyChecker()
        html = checker.render_file_alert_html(
            error_ids='error1',
            warning_ids='warning1')

        self.assertEqual(html, '<html>result</html>')

    def test_render_file_alert_html_without_template(self):
        """Testing BaseCodeSafetyChecker.render_file_alert_html without
        template
        """
        self.assertIsNone(self.checker.render_file_alert_html(
            error_ids='error1',
            warning_ids='warning1'))

    def test_get_file_alert_context_data(self):
        """Testing BaseCodeSafetyChecker.get_file_alert_context_data"""
        self.assertEqual(
            self.checker.get_file_alert_context_data(
                error_ids=['error1', 'error2'],
                warning_ids=['warning1', 'warning2']),
            {
                'error_ids': ['error1', 'error2'],
                'error_labels': ['Error #1', 'error2'],
                'warning_ids': ['warning1', 'warning2'],
                'warning_labels': ['Warning #1', 'warning2'],
            })



"""Unit tests for reviewboard.admin.views.dashboard."""

from kgb import SpyAgency

from reviewboard.admin.views import logger
from reviewboard.admin.widgets import BaseAdminWidget, admin_widgets_registry
from reviewboard.testing.testcase import TestCase


class MyWidget(BaseAdminWidget):
    widget_id = 'my-widget'
    name = 'My Widget'
    css_classes = 'test-c-my-widget -is-large'

    def get_js_model_attrs(self, request):
        return {
            'test_attr': 'test-value',
        }

    def get_js_model_options(self, request):
        return {
            'test_option': 'test-value',
        }

    def get_js_view_options(self, request):
        return {
            'test_option': 'test-value',
        }


class HiddenWidget(BaseAdminWidget):
    widget_id = 'hidden-widget'
    name = 'Hidden Widget'

    def can_render(self, request):
        return False


class AdminDashboardViewTests(SpyAgency, TestCase):
    """Unit tests for reviewboard.admin.views.admin_dashboard_view."""

    fixtures = ['test_users']

    def test_get(self):
        """Testing admin_dashboard_view"""
        admin_widgets_registry.register(MyWidget)

        try:
            self.client.login(username='admin', password='admin')
            response = self.client.get('/admin/')
            self.assertEqual(response.status_code, 200)

            self.assertIn('page_model_attrs', response.context)
            widgets = response.context['page_model_attrs']['widgetsData']

            self.assertEqual(len(widgets), len(admin_widgets_registry))

            self.assertEqual(
                widgets[-1],
                {
                    'id': 'my-widget',
                    'domID': 'admin-widget-my-widget',
                    'viewClass': 'RB.Admin.WidgetView',
                    'modelClass': 'RB.Admin.Widget',
                    'viewOptions': {
                        'test_option': 'test-value',
                    },
                    'modelAttrs': {
                        'test_attr': 'test-value',
                    },
                    'modelOptions': {
                        'test_option': 'test-value',
                    },
                })

            self.assertIn(
                b'<div class="rb-c-admin-widget test-c-my-widget -is-large"'
                b' id="admin-widget-my-widget">',
                response.content)
        finally:
            admin_widgets_registry.unregister(MyWidget)

    def test_get_with_widget_can_render_false(self):
        """Testing admin_dashboard_view with widget.can_render() == False"""
        admin_widgets_registry.register(HiddenWidget)

        try:
            self.client.login(username='admin', password='admin')
            response = self.client.get('/admin/')
            self.assertEqual(response.status_code, 200)

            self.assertIn('page_model_attrs', response.context)
            widgets = response.context['page_model_attrs']['widgetsData']

            self.assertEqual(len(widgets), len(admin_widgets_registry) - 1)

            self.assertNotEqual(widgets[-1]['id'], 'hidden-widget')
        finally:
            admin_widgets_registry.unregister(HiddenWidget)

    def test_get_with_broken_widget_init(self):
        """Testing admin_dashboard_view with broken widget.__init__"""
        error_msg = '__init__ broke'

        class BrokenWidget(BaseAdminWidget):
            widget_id = 'broken-widget'

            def __init__(self):
                raise Exception(error_msg)

        self._test_broken_widget(BrokenWidget, error_msg)

    def test_get_with_broken_widget_can_render(self):
        """Testing admin_dashboard_view with broken widget.can_render"""
        error_msg = 'can_render broke'

        class BrokenWidget(BaseAdminWidget):
            widget_id = 'broken-widget'

            def can_render(self, request):
                raise Exception(error_msg)

        self._test_broken_widget(BrokenWidget, error_msg)

    def test_get_with_broken_widget_get_js_view_options(self):
        """Testing admin_dashboard_view with broken widget.get_js_view_options
        """
        error_msg = 'get_js_view_options broke'

        class BrokenWidget(BaseAdminWidget):
            widget_id = 'broken-widget'

            def get_js_view_options(self, request):
                raise Exception(error_msg)

        self._test_broken_widget(BrokenWidget, error_msg)

    def test_get_with_broken_widget_get_js_model_attrs(self):
        """Testing admin_dashboard_view with broken widget.get_js_model_attrs
        """
        error_msg = 'get_js_model_attrs broke'

        class BrokenWidget(BaseAdminWidget):
            widget_id = 'broken-widget'

            def get_js_model_attrs(self, request):
                raise Exception(error_msg)

        self._test_broken_widget(BrokenWidget, error_msg)

    def test_get_with_broken_widget_get_js_model_options(self):
        """Testing admin_dashboard_view with broken widget.get_js_model_options
        """
        error_msg = 'get_js_model_options broke'

        class BrokenWidget(BaseAdminWidget):
            widget_id = 'broken-widget'

            def get_js_model_options(self, request):
                raise Exception(error_msg)

        self._test_broken_widget(BrokenWidget, error_msg)

    def test_get_with_broken_widget_render(self):
        """Testing admin_dashboard_view with broken widget.render"""
        error_msg = 'render broke'

        class BrokenWidget(BaseAdminWidget):
            widget_id = 'broken-widget'

            def render(self, request):
                raise Exception(error_msg)

        self._test_broken_widget(BrokenWidget, error_msg)

    def _test_broken_widget(self, widget_cls, expected_msg):
        """Test that a broken widget doesn't break the dashboard view.

        Args:
            widget_cls (type):
                The broken widget to register and test against.

            expected_msg (unicode, optional):
                The expected error message raised by the broken method.
        """
        admin_widgets_registry.register(widget_cls)
        self.spy_on(logger.exception)

        try:
            self.client.login(username='admin', password='admin')
            response = self.client.get('/admin/')
            self.assertEqual(response.status_code, 200)

            self.assertIn('page_model_attrs', response.context)
            widgets = response.context['page_model_attrs']['widgetsData']

            self.assertEqual(len(widgets), len(admin_widgets_registry) - 1)
            self.assertNotEqual(widgets[-1]['id'], widget_cls.widget_id)
            self.assertNotIn(
                b'id="admin-widget-%s"' % widget_cls.widget_id.encode('utf-8'),
                response.content)

            spy_call = logger.exception.last_call
            self.assertEqual(spy_call.args[0],
                             'Error setting up administration widget %r: %s')
            self.assertEqual(spy_call.args[1], widget_cls)
            self.assertIsInstance(spy_call.args[2], Exception)
            self.assertEqual(str(spy_call.args[2]), expected_msg)
        finally:
            admin_widgets_registry.unregister(widget_cls)



"""Unit tests for reviewboard.extensions.hooks.EmailHook and subclasses."""

from django.contrib.auth.models import User
from django.core import mail
from djblets.mail.utils import build_email_address_for_user
from kgb import SpyAgency

from reviewboard.extensions.hooks import (EmailHook,
                                          ReviewPublishedEmailHook,
                                          ReviewReplyPublishedEmailHook,
                                          ReviewRequestClosedEmailHook,
                                          ReviewRequestPublishedEmailHook)
from reviewboard.extensions.tests.testcases import BaseExtensionHookTestCase
from reviewboard.reviews.models import ReviewRequest
from reviewboard.reviews.signals import (review_request_published,
                                         review_published,
                                         reply_published,
                                         review_request_closed)


class EmailHookTests(SpyAgency, BaseExtensionHookTestCase):
    """Testing the e-mail recipient filtering capacity of EmailHooks."""

    fixtures = ['test_users']

    def setUp(self):
        super(EmailHookTests, self).setUp()

        mail.outbox = []

    def test_review_request_published_email_hook(self):
        """Testing ReviewRequestPublishedEmailHook"""
        class DummyHook(ReviewRequestPublishedEmailHook):
            def get_to_field(self, to_field, review_request, user):
                return set([user])

            def get_cc_field(self, cc_field, review_request, user):
                return set([user])

        hook = DummyHook(self.extension)

        self.spy_on(hook.get_to_field)
        self.spy_on(hook.get_cc_field)

        review_request = self.create_review_request()
        admin = User.objects.get(username='admin')

        call_kwargs = {
            'user': admin,
            'review_request': review_request,
        }

        with self.siteconfig_settings({'mail_send_review_mail': True}):
            review_request.publish(admin)

        self.assertEqual(len(mail.outbox), 1)
        self.assertEqual(mail.outbox[0].to,
                         [build_email_address_for_user(admin)])
        self.assertTrue(hook.get_to_field.called)
        self.assertTrue(hook.get_to_field.called_with(**call_kwargs))
        self.assertTrue(hook.get_cc_field.called)
        self.assertTrue(hook.get_cc_field.called_with(**call_kwargs))

    def test_review_published_email_hook(self):
        """Testing ReviewPublishedEmailHook"""
        class DummyHook(ReviewPublishedEmailHook):
            def get_to_field(self, to_field, review, user, review_request,
                             to_owner_only):
                return set([user])

            def get_cc_field(self, cc_field, review, user, review_request,
                             to_owner_only):
                return set([user])

        hook = DummyHook(self.extension)

        self.spy_on(hook.get_to_field)
        self.spy_on(hook.get_cc_field)

        admin = User.objects.get(username='admin')
        review_request = self.create_review_request(public=True)
        review = self.create_review(review_request)

        call_kwargs = {
            'user': admin,
            'review_request': review_request,
            'review': review,
            'to_owner_only': False,
        }

        with self.siteconfig_settings({'mail_send_review_mail': True}):
            review.publish(admin)

        self.assertEqual(len(mail.outbox), 1)
        self.assertEqual(mail.outbox[0].to,
                         [build_email_address_for_user(admin)])
        self.assertTrue(hook.get_to_field.called)
        self.assertTrue(hook.get_to_field.called_with(**call_kwargs))
        self.assertTrue(hook.get_cc_field.called)
        self.assertTrue(hook.get_cc_field.called_with(**call_kwargs))

    def test_review_reply_published_email_hook(self):
        """Testing ReviewReplyPublishedEmailHook"""
        class DummyHook(ReviewReplyPublishedEmailHook):
            def get_to_field(self, to_field, reply, user, review,
                             review_request):
                return set([user])

            def get_cc_field(self, cc_field, reply, user, review,
                             review_request):
                return set([user])

        hook = DummyHook(self.extension)

        self.spy_on(hook.get_to_field)
        self.spy_on(hook.get_cc_field)

        admin = User.objects.get(username='admin')
        review_request = self.create_review_request(public=True)
        review = self.create_review(review_request)
        reply = self.create_reply(review)

        call_kwargs = {
            'user': admin,
            'review_request': review_request,
            'review': review,
            'reply': reply,
        }

        with self.siteconfig_settings({'mail_send_review_mail': True}):
            reply.publish(admin)

        self.assertEqual(len(mail.outbox), 1)
        self.assertTrue(hook.get_to_field.called)
        self.assertTrue(hook.get_to_field.called_with(**call_kwargs))
        self.assertTrue(hook.get_cc_field.called)
        self.assertTrue(hook.get_cc_field.called_with(**call_kwargs))

    def test_review_request_closed_email_hook_submitted(self):
        """Testing ReviewRequestClosedEmailHook for a review request being
        submitted
        """
        class DummyHook(ReviewRequestClosedEmailHook):
            def get_to_field(self, to_field, review_request, user, close_type):
                return set([user])

            def get_cc_field(self, cc_field, review_request, user, close_type):
                return set([user])

        hook = DummyHook(self.extension)

        self.spy_on(hook.get_to_field)
        self.spy_on(hook.get_cc_field)

        admin = User.objects.get(username='admin')
        review_request = self.create_review_request(public=True)

        call_kwargs = {
            'user': admin,
            'review_request': review_request,
            'close_type': ReviewRequest.SUBMITTED,
        }

        with self.siteconfig_settings({'mail_send_review_close_mail': True}):
            review_request.close(ReviewRequest.SUBMITTED, admin)

        self.assertEqual(len(mail.outbox), 1)
        self.assertTrue(hook.get_to_field.called)
        self.assertTrue(hook.get_to_field.called_with(**call_kwargs))
        self.assertTrue(hook.get_cc_field.called)
        self.assertTrue(hook.get_cc_field.called_with(**call_kwargs))

    def test_review_request_closed_email_hook_discarded(self):
        """Testing ReviewRequestClosedEmailHook for a review request being
        discarded
        """
        class DummyHook(ReviewRequestClosedEmailHook):
            def get_to_field(self, to_field, review_request, user, close_type):
                return set([user])

            def get_cc_field(self, cc_field, review_request, user, close_type):
                return set([user])

        hook = DummyHook(self.extension)

        self.spy_on(hook.get_to_field)
        self.spy_on(hook.get_cc_field)

        admin = User.objects.get(username='admin')
        review_request = self.create_review_request(public=True)

        call_kwargs = {
            'user': admin,
            'review_request': review_request,
            'close_type': ReviewRequest.DISCARDED,
        }

        with self.siteconfig_settings({'mail_send_review_close_mail': True}):
            review_request.close(ReviewRequest.DISCARDED, admin)

        self.assertEqual(len(mail.outbox), 1)
        self.assertTrue(hook.get_to_field.called)
        self.assertTrue(hook.get_to_field.called_with(**call_kwargs))
        self.assertTrue(hook.get_cc_field.called)
        self.assertTrue(hook.get_cc_field.called_with(**call_kwargs))

    def test_generic_hook(self):
        """Testing EmailHook connects to all signals"""
        hook = EmailHook(self.extension,
                         signals=[
                             review_request_published,
                             review_published,
                             reply_published,
                             review_request_closed,
                         ])

        self.spy_on(hook.get_to_field)
        self.spy_on(hook.get_cc_field)

        user = User.objects.create_user(username='testuser')
        review_request = self.create_review_request(public=True,
                                                    target_people=[user])
        review = self.create_review(review_request)
        reply = self.create_reply(review)

        siteconfig_settings = {
            'mail_send_review_mail': True,
            'mail_send_review_close_mail': True,
        }

        with self.siteconfig_settings(siteconfig_settings):
            self.assertEqual(len(mail.outbox), 0)

            review.publish()
            call_kwargs = {
                'user': review.user,
                'review': review,
                'review_request': review_request,
                'to_owner_only': False,
            }

            self.assertEqual(len(mail.outbox), 1)
            self.assertEqual(len(hook.get_to_field.spy.calls), 1)
            self.assertEqual(len(hook.get_cc_field.spy.calls), 1)
            self.assertEqual(hook.get_to_field.spy.calls[-1].kwargs,
                             call_kwargs)
            self.assertEqual(hook.get_cc_field.spy.calls[-1].kwargs,
                             call_kwargs)

            reply.publish(reply.user)

            call_kwargs.pop('to_owner_only')
            call_kwargs['reply'] = reply
            call_kwargs['user'] = reply.user

            self.assertEqual(len(mail.outbox), 2)
            self.assertEqual(len(hook.get_to_field.spy.calls), 2)
            self.assertEqual(len(hook.get_cc_field.spy.calls), 2)
            self.assertEqual(hook.get_to_field.spy.calls[-1].kwargs,
                             call_kwargs)
            self.assertEqual(hook.get_cc_field.spy.calls[-1].kwargs,
                             call_kwargs)

            review_request.close(ReviewRequest.DISCARDED)
            call_kwargs = {
                'review_request': review_request,
                'user': review_request.submitter,
                'close_type': ReviewRequest.DISCARDED,
            }

            self.assertEqual(len(mail.outbox), 3)
            self.assertEqual(len(hook.get_to_field.spy.calls), 3)
            self.assertEqual(len(hook.get_cc_field.spy.calls), 3)
            self.assertEqual(hook.get_to_field.spy.calls[-1].kwargs,
                             call_kwargs)
            self.assertEqual(hook.get_cc_field.spy.calls[-1].kwargs,
                             call_kwargs)

            review_request.reopen()
            review_request.publish(review_request.submitter)
            call_kwargs = {
                'review_request': review_request,
                'user': review_request.submitter,
            }

            self.assertEqual(len(mail.outbox), 4)
            self.assertEqual(len(hook.get_to_field.spy.calls), 4)
            self.assertEqual(len(hook.get_cc_field.spy.calls), 4)
            self.assertEqual(hook.get_to_field.spy.calls[-1].kwargs,
                             call_kwargs)
            self.assertEqual(hook.get_cc_field.spy.calls[-1].kwargs,
                             call_kwargs)

            review_request.close(ReviewRequest.SUBMITTED)
            call_kwargs['close_type'] = ReviewRequest.SUBMITTED

            self.assertEqual(len(mail.outbox), 5)
            self.assertEqual(len(hook.get_to_field.spy.calls), 5)
            self.assertEqual(len(hook.get_cc_field.spy.calls), 5)
            self.assertEqual(hook.get_to_field.spy.calls[-1].kwargs,
                             call_kwargs)
            self.assertEqual(hook.get_cc_field.spy.calls[-1].kwargs,
                             call_kwargs)



"""Base class for authentication backends."""

from __future__ import annotations

import re
from typing import Optional, TYPE_CHECKING, Type, Union

from django.contrib.auth.models import User
from djblets.db.query import get_object_or_none

if TYPE_CHECKING:
    from django.db.models import Q
    from django.http import HttpRequest
    from djblets.siteconfig.forms import SiteSettingsForm
    from typelets.django.strings import StrOrPromise


class BaseAuthBackend:
    """Base class for a Review Board authentication backend."""

    #: The unique ID for the authentication backend.
    #:
    #: Type:
    #:     str
    backend_id: Optional[str] = None

    #: The display name for the authentication backend.
    #:
    #: This will be shown in the list of backends in the administration UI.
    #:
    #: Type:
    #:     str
    name: Optional[StrOrPromise] = None

    #: The form class used for authentication settings.
    #:
    #: This must be a subclass of
    #: :py:class:`~djblets.siteconfig.forms.SiteSettingsForm`.
    #:
    #: Type:
    #:     type
    settings_form: Optional[Type[SiteSettingsForm]] = None

    #: Whether this backend supports registering new users.
    #:
    #: Type:
    #:     bool
    supports_registration: bool = False

    #: Whether this backend supports changing the user's full name.
    #:
    #: Type:
    #:     bool
    supports_change_name: bool = False

    #: Whether this backend supports changing the user's e-mail address.
    #:
    #: Type:
    #:     bool
    supports_change_email: bool = False

    #: Whether this backend supports changing the user's password.
    #:
    #: Type:
    #:     bool
    supports_change_password: bool = False

    #: Authentication instructions to display above the Login form.
    #:
    #: Type:
    #:     str
    login_instructions: Optional[StrOrPromise] = None

    #: A regex for matching invalid characters in usernames.
    #:
    #: Type:
    #:     re.Pattern
    INVALID_USERNAME_CHAR_REGEX: re.Pattern = re.compile(r'[^\w.@+-]')

    def authenticate(
        self,
        request: Optional[HttpRequest] = None,
        **credentials,
    ) -> Optional[User]:
        """Authenticate a user.

        This will authenticate a user identified by the provided credentials.
        object, or None.

        This must be implemented by subclasses.

        Version Changed:
            6.0:
            ``request`` is now optional.

        Version Changed:
            4.0:
            The ``request`` argument is now mandatory as the first positional
            argument, as per requirements in Django.

        Args:
            request (django.http.HttpRequest):
                The HTTP request from the caller. This may be ``None``.

            **credentials (dict):
                The credentials passed to the backend. This will often
                contain ``username`` and ``password`` keys.

        Returns:
            django.contrib.auth.models.User:
            The authenticated user, or ``None`` if the user could not be
            authenticated for any reason.
        """
        raise NotImplementedError

    def get_or_create_user(
        self,
        username: str,
        request: Optional[HttpRequest] = None,
    ) -> Optional[User]:
        """Return an existing user or create one if it doesn't exist.

        This does not authenticate the user.

        If the user does not exist in the database, but does in the backend,
        its information will be stored in the database for later lookup.
        Subclasses can impose restrictions on this.

        This must be implemented by subclasses.

        Args:
            username (str):
                The username to fetch or create.

            request (django.http.HttpRequest, optional):
                The HTTP request from the client.

        Returns:
            django.contrib.auth.models.User:
            The resulting user, or ``None`` if one could not be found.
        """
        raise NotImplementedError

    def get_user(
        self,
        user_id: Union[int, str],
    ) -> Optional[User]:
        """Return an existing user given a numeric user ID.

        Args:
            user_id (int or str):
                The ID of the user to retrieve.

        Returns:
            django.contrib.auth.models.User:
            The resulting user, or ``None`` if one could not be found.
        """
        return get_object_or_none(User, pk=user_id)

    def update_password(
        self,
        user: User,
        password: str,
    ) -> None:
        """Update a user's password on the backend.

        Authentication backends can override this to update the password
        on the backend. By default, this is not supported.

        Callers should only call this after checking whether
        :py:attr:`supports_change_password` is ``True``.

        Args:
            user (django.contrib.auth.models.User):
                The user whose password will be changed.

            password (str):
                The new password.

        Raises:
            NotImplementedError:
                The backend does not support changing passwords.
        """
        raise NotImplementedError

    def update_name(
        self,
        user: User,
    ) -> None:
        """Update a user's full name on the backend.

        Authentication backends can override this to update the name on the
        backend based on the
        :py:attr:`~django.contrib.auth.models.User.first_name` and
        :py:attr:`~django.contrib.auth.models.User.last_name` values in
        ``user``. By default, this will do nothing.

        Callers should only call this after checking whether
        :py:attr:`supports_change_name` is ``True``.

        Args:
            user (django.contrib.auth.models.User):
                The user whose full name will be changed.
        """
        pass

    def update_email(
        self,
        user: User,
    ) -> None:
        """Update a user's e-mail address on the backend.

        Authentication backends can override this to update the e-mail
        address on the backend based on the
        :py:attr:`~django.contrib.auth.models.User.email` value in ``user``.
        NBy default, this will do nothing.

        Callers should only call this after checking whether
        :py:attr:`supports_change_email` is ``True``.

        Args:
            user (django.contrib.auth.models.User):
                The user whose full name will be changed.
        """
        pass

    def populate_users(
        self,
        query: str,
        request: Optional[HttpRequest] = None,
        **kwargs,
    ) -> None:
        """Populate users from the backend into the database based on a query.

        Authentication backends can override this to add users stored on the
        backend into the local database, based on a query string representing
        a full or partial first name, last name, or username. Each result
        that's found based on the query should be stored in the database by
        the backend, generally using the same creation logic as in
        :py:meth:`get_or_create_user`.

        Callers should use this when they need to look up all available users
        from a backend. After calling this, they should look up the results
        from the database.

        By default, this does nothing.

        Version Changed:
            6.0:
            ``request`` is now optional.

        Args:
            query (str):
                A search query for matching users. This will match the entirety
                or prefix of a username. It's expected that the match will be
                case-insensitive.

            request (django.http.HttpRequest):
                The HTTP request from the client.

            **kwargs (dict):
                Extra positional arguments, for future use.

        Raises:
            reviewboard.accounts.errors.UserQueryError:
                There was an error processing the query or looking up users.
                Details will be in the error message.
        """
        pass

    def build_search_users_query(
        self,
        query: str,
        request: Optional[HttpRequest] = None,
        **kwargs,
    ) -> Optional[Q]:
        """Build a query for searching users in the database.

        This allows backends to construct specialized search queries (
        for use in :py:meth:`QuerySet.filter()
        <django.db.models.query.QuerySet.filter>` when searching for users
        via the :ref:`webapi2.0-user-list-resource`.

        By default, this will return ``None``.

        Version Changed:
            6.0:
            ``request`` is now optional.

        Args:
            query (str):
                A search query for matching users. This will match the entirety
                or prefix of a username. It's expected that the match will be
                case-insensitive.

            request (django.http.HttpRequest):
                The HTTP request from the client.

            **kwargs (dict):
                Extra positional arguments, for future use.

        Returns:
            django.db.models.Q:
            The resulting query for the queryset, or ``None`` to use the
            query for the next available backend (eventually defaulting to
            the standard search query).
        """
        return None



"""Unit tests for reviewboard.attachments.mimetypes.

Version Added:
    7.0.3
"""

from __future__ import annotations

import os
import subprocess
from unittest.mock import patch

import kgb
from django.conf import settings
from django.core.cache import cache
from django.core.files.uploadedfile import SimpleUploadedFile
from djblets.cache.backend import make_cache_key
from djblets.util.filesystem import is_exe_in_path

from reviewboard.attachments.mimetypes import (guess_mimetype,
                                               is_mimetype_supported)
from reviewboard.testing import TestCase
from reviewboard.webapi.server_info import get_capabilities


class GuessMimetypeTests(kgb.SpyAgency, TestCase):
    """Unit tests for guess_mimetype.

    Version Added:
        7.0.3
    """

    def test_with_text(self) -> None:
        """Testing guess_mimetype with text file"""
        file = SimpleUploadedFile('test.txt', b'this is a text\n')

        self.assertEqual(guess_mimetype(file), 'text/plain')
        self.assertEqual(file.tell(), 0)

    def test_with_image(self) -> None:
        """Testing guess_mimetype with image file"""
        filename = os.path.join(settings.STATIC_ROOT,
                                'rb', 'images', 'logo.png')

        with open(filename, 'rb') as fp:
            file = SimpleUploadedFile(fp.name, fp.read())

        self.assertEqual(guess_mimetype(file), 'image/png')
        self.assertEqual(file.tell(), 0)

    def test_with_pdf(self) -> None:
        """Testing guess_mimetype with PDF"""
        file = SimpleUploadedFile('test.pdf', b'\x25\x50\x44\x46\x2d')

        self.assertEqual(guess_mimetype(file), 'application/pdf')
        self.assertEqual(file.tell(), 0)

    def test_with_unknown(self) -> None:
        """Testing guess_mimetype with unknown file type"""
        file = SimpleUploadedFile('image.png', b'\0\1\2\3\4\5')

        self.assertEqual(guess_mimetype(file), 'application/octet-stream')
        self.assertEqual(file.tell(), 0)

    def test_with_large_file(self) -> None:
        """Testing guess_mimetype with large file"""
        # We'll pick a big payload size that should go beyond the read length
        # from `file`, causing the pipe to close early.
        #
        # Not that not all Python subprocess.Popen implementations will crash
        # when trying to close a pipe more than once.
        file = SimpleUploadedFile(
            'test.bin',
            b'\x25\x50\x44\x46\x2d' + (b'\0' * 1024 * 1024 * 5))

        self.assertEqual(guess_mimetype(file), 'application/pdf')
        self.assertEqual(file.tell(), 0)

    def test_with_exit_error(self) -> None:
        """Testing guess_mimetype with non-0 exit code"""
        self.spy_on(subprocess.Popen.wait,
                    owner=subprocess.Popen,
                    op=kgb.SpyOpReturn(1))

        file = SimpleUploadedFile('test.pdf', b'\x25\x50\x44\x46\x2d')

        self.assertEqual(guess_mimetype(file), 'application/octet-stream')
        self.assertEqual(file.tell(), 0)

    def test_with_file_exe_not_found(self) -> None:
        """Testing guess_mimetype with file executable not found"""
        self.spy_on(is_exe_in_path, op=kgb.SpyOpReturn(False))

        file = SimpleUploadedFile('test.pdf', b'\x25\x50\x44\x46\x2d')

        self.assertEqual(guess_mimetype(file), 'application/octet-stream')
        self.assertEqual(file.tell(), 0)


class IsMimetypeSupportedTests(kgb.SpyAgency, TestCase):
    """Unit tests for is_mimetype_supported.

    Version Added:
        7.0.4
    """

    def test_with_text(self) -> None:
        """Testing is_mimetype_supported with text file"""
        self.assertTrue(is_mimetype_supported('text/plain'))

    def test_with_image(self) -> None:
        """Testing is_mimetype_supported with image file"""
        self.assertTrue(is_mimetype_supported('image/png'))

    def test_with_pdf(self) -> None:
        """Testing is_mimetype_supported with PDF"""
        self.assertTrue(is_mimetype_supported('application/pdf'))

    def test_with_unknown(self) -> None:
        """Testing is_mimetype_supported with unknown file type"""
        self.assertFalse(is_mimetype_supported('application/octet-stream'))

    def test_cache_supported(self) -> None:
        """Testing is_mimetype_supported sets the cache for supported files"""
        self.spy_on(cache.set)
        key = make_cache_key('review-uis:supported-mimetypes:text/plain')

        self.assertIsNone(cache.get(key))
        self.assertTrue(is_mimetype_supported('text/plain'))
        self.assertTrue(cache.get(key))
        self.assertTrue(is_mimetype_supported('text/plain'))
        self.assertSpyCallCount(cache.set, 1)

    def test_cache_unsupported(self) -> None:
        """Testing is_mimetype_supported sets the cache for unsupported files
        """
        self.spy_on(cache.set)
        key = make_cache_key(
            'review-uis:supported-mimetypes:application/octet-stream')

        self.assertIsNone(cache.get(key))
        self.assertFalse(is_mimetype_supported('application/octet-stream'))
        self.assertFalse(cache.get(key))
        self.assertFalse(is_mimetype_supported('application/octet-stream'))
        self.assertSpyCallCount(cache.set, 1)

    @patch('reviewboard.attachments.mimetypes._supported_mimetypes', [])
    def test_cache_supported_list(self) -> None:
        """Testing is_mimetype_supported uses the in-memory list of supported
        mimetypes if it has been set
        """
        cache.clear()
        self.spy_on(get_capabilities)

        self.assertTrue(is_mimetype_supported('text/plain'))
        self.assertTrue(is_mimetype_supported('application/pdf'))
        self.assertSpyCallCount(get_capabilities, 1)



"""Add Group.is_default_group.

Version Added:
    2.5
"""

from django_evolution.mutations import AddField
from django.db import models


MUTATIONS = [
    AddField('Group', 'is_default_group', models.BooleanField, initial=False)
]



"""Unit tests for reviewboard.extensions.hooks.ActionHook and subclasses."""

from typing import Any, Dict, Type

from django.template import Context, Template

from reviewboard.actions import (AttachmentPoint,
                                 BaseAction,
                                 BaseMenuAction,
                                 actions_registry)
from reviewboard.actions.base import ActionPlacement
from reviewboard.deprecation import RemovedInReviewBoard90Warning
from reviewboard.extensions.hooks import (ActionHook,
                                          DiffViewerActionHook,
                                          HeaderActionHook,
                                          HeaderDropdownActionHook,
                                          HideActionHook,
                                          ReviewRequestActionHook,
                                          ReviewRequestDropdownActionHook)
from reviewboard.extensions.hooks.actions import BaseReviewRequestActionHook
from reviewboard.extensions.tests.testcases import BaseExtensionHookTestCase
from reviewboard.reviews.actions import (BaseReviewRequestAction,
                                         BaseReviewRequestMenuAction)


class ActionHookTests(BaseExtensionHookTestCase):
    """Tests for modern ActionHook usage."""

    fixtures = ['test_users']

    class _TestAction(BaseAction):
        action_id = 'test'
        label = 'Test Action'

    class _TestMenuAction(BaseMenuAction):
        action_id = 'test-menu'
        label = 'Test Menu'

    class _TestMenuInstance(BaseAction):
        action_id = 'test-menu-item'
        label = 'Test Menu Item'

        placements = [
            ActionPlacement(attachment=AttachmentPoint.REVIEW_REQUEST,
                            parent_id='test-menu'),
        ]

    def tearDown(self) -> None:
        """Tear down the test case."""
        super().tearDown()
        actions_registry.reset()

    def test_action_hook(self) -> None:
        """Testing ActionHook registration"""
        test_action = self._TestAction()
        test_menu_action = self._TestMenuAction()
        test_menu_instance = self._TestMenuInstance()

        test_menu_action_placement = \
            test_menu_action.get_placement('review-request')
        test_menu_instance_placement = \
            test_menu_instance.get_placement('review-request')

        hook = ActionHook(extension=self.extension, actions=[
            test_action,
            test_menu_action,
            test_menu_instance,
        ])

        try:
            self.assertEqual(
                actions_registry.get_action('test'),
                test_action)
            self.assertEqual(
                actions_registry.get_action('test-menu'),
                test_menu_action)
            self.assertEqual(
                actions_registry.get_action('test-menu-item'),
                test_menu_instance)

            self.assertEqual(test_menu_instance_placement.parent_action,
                             test_menu_action)
            self.assertEqual(test_menu_action_placement.child_actions,
                             [test_menu_instance])
        finally:
            hook.disable_hook()

        self.assertIsNone(actions_registry.get_action('test'))
        self.assertIsNone(actions_registry.get_action('test-menu'))
        self.assertIsNone(actions_registry.get_action('test-menu-item'))
        self.assertIsNone(test_menu_instance_placement.parent_action)
        self.assertEqual(test_menu_action_placement.child_actions, [])


class LegacyActionHookTests(BaseExtensionHookTestCase):
    """Tests for the legacy action hooks in reviewboard.extensions.hooks."""

    fixtures = ['test_users']

    class _TestAction(BaseReviewRequestAction):
        action_id = 'test-action'
        label = 'Test Action'

    class _TestMenuAction(BaseReviewRequestMenuAction):
        action_id = 'test-menu-instance'
        label = 'Menu Instance'

    def setUp(self) -> None:
        """Set up the test case."""
        super().setUp()

        # Pre-populate this, so population side effects (e.g., warnings)
        # won't interfere with any tests.
        actions_registry.populate()

    def tearDown(self) -> None:
        """Tear down the test case."""
        super().tearDown()

        actions_registry.reset()

    def test_review_request_action_hook(self) -> None:
        """Testing ReviewRequestActionHook renders on a review request page but
        not on a file attachment or a diff viewer page
        """
        deprecation_message = (
            'ReviewRequestActionHook is deprecated and will be removed in '
            'Review Board 9. Your extension '
            '"reviewboard.extensions.tests.testcases.DummyExtension" '
            'will need to be updated to derive actions from '
            'reviewboard.actions.BaseAction and use ActionHook.'
        )

        with self.assertWarns(RemovedInReviewBoard90Warning,
                              deprecation_message):
            self._test_base_review_request_action_hook(
                'review-request-detail', ReviewRequestActionHook, True)

        with self.assertWarns(RemovedInReviewBoard90Warning,
                              deprecation_message):
            self._test_base_review_request_action_hook(
                'file-attachment', ReviewRequestActionHook, False)

        with self.assertWarns(RemovedInReviewBoard90Warning,
                              deprecation_message):
            self._test_base_review_request_action_hook(
                'view-diff', ReviewRequestActionHook, False)

    def test_diffviewer_action_hook(self) -> None:
        """Testing DiffViewerActionHook renders on a diff viewer page but not
        on a review request page or a file attachment page
        """
        deprecation_message = (
            'DiffViewerActionHook is deprecated and will be removed in '
            'Review Board 9. Your extension '
            '"reviewboard.extensions.tests.testcases.DummyExtension" '
            'will need to be updated to derive actions from '
            'reviewboard.actions.BaseAction and use ActionHook.'
        )

        with self.assertWarns(RemovedInReviewBoard90Warning,
                              deprecation_message):
            self._test_base_review_request_action_hook(
                'review-request-detail', DiffViewerActionHook, False)

        with self.assertWarns(RemovedInReviewBoard90Warning,
                              deprecation_message):
            self._test_base_review_request_action_hook(
                'file-attachment', DiffViewerActionHook, False)

        with self.assertWarns(RemovedInReviewBoard90Warning,
                              deprecation_message):
            self._test_base_review_request_action_hook(
                'view-diff', DiffViewerActionHook, True)

    def test_review_request_dropdown_action_hook(self) -> None:
        """Testing ReviewRequestDropdownActionHook renders on a review request
        page but not on a file attachment or a diff viewer page
        """
        deprecation_message = (
            'ReviewRequestDropdownActionHook is deprecated and will be '
            'removed in Review Board 9. Your extension '
            '"reviewboard.extensions.tests.testcases.DummyExtension" '
            'will need to be updated to derive actions from '
            'reviewboard.actions.BaseAction and use ActionHook.'
        )

        with self.assertWarns(RemovedInReviewBoard90Warning,
                              deprecation_message):
            self._test_review_request_dropdown_action_hook(
                'review-request-detail', ReviewRequestDropdownActionHook)

        with self.assertWarns(RemovedInReviewBoard90Warning,
                              deprecation_message):
            self._test_review_request_dropdown_action_hook(
                'file-attachment', ReviewRequestDropdownActionHook)

        with self.assertWarns(RemovedInReviewBoard90Warning,
                              deprecation_message):
            self._test_review_request_dropdown_action_hook(
                'view-diff', ReviewRequestDropdownActionHook)

    def test_action_hook_init_raises_key_error(self) -> None:
        """Testing ActionHook.__init__ with raised KeyError"""
        missing_url_action = {
            'id': 'missing-url-action',
            'label': 'This action dict is missing a mandatory URL key.',
        }
        missing_key = 'url'
        error_message = ('Action dictionaries require a %s key'
                         % repr(missing_key))
        action_hook_classes = [
            BaseReviewRequestActionHook,
            ReviewRequestActionHook,
            DiffViewerActionHook,
        ]

        for hook_cls in action_hook_classes:
            with self.assertRaisesMessage(KeyError, error_message):
                hook_cls(extension=self.extension, actions=[
                    missing_url_action,
                ])

    def test_action_hook_init_raises_value_error(self) -> None:
        """Testing ActionHook.__init__ with raised ValueError"""
        unsupported_type_action = [{
            'id': 'unsupported-type-action',
            'label': 'This action is a list, which is an unsupported type.',
            'url': '#',
        }]
        error_message = 'Action definitions must be dictionaries.'
        action_hook_classes = [
            BaseReviewRequestActionHook,
            ReviewRequestActionHook,
            DiffViewerActionHook,
            ReviewRequestDropdownActionHook,
        ]

        for hook_cls in action_hook_classes:
            with self.assertRaisesMessage(ValueError, error_message):
                hook_cls(extension=self.extension, actions=[
                    unsupported_type_action,
                ])

    def test_dropdown_action_hook_init_raises_key_error(self) -> None:
        """Testing ReviewRequestDropdownActionHook.__init__ with raiseed
        KeyError
        """
        missing_items_menu_action = {
            'id': 'missing-items-menu-action',
            'label': 'This menu action dict is missing a mandatory items key.',
        }

        with self.assertRaisesMessage(KeyError, 'items'):
            ReviewRequestDropdownActionHook(extension=self.extension, actions=[
                missing_items_menu_action,
            ])

    def test_header_hooks(self) -> None:
        """Testing HeaderActionHook"""
        action = {
            'id': 'test',
            'image': 'test-image',
            'image_height': 42,
            'image_width': 42,
            'label': 'Test Action',
            'url': 'foo-url',
        }

        deprecation_message = (
            'HeaderActionHook is deprecated and will be removed in Review '
            'Board 9. Your extension '
            '"reviewboard.extensions.tests.testcases.DummyExtension" '
            'will need to be updated to derive actions from '
            'reviewboard.actions.BaseAction and use ActionHook.'
        )

        with self.assertWarns(RemovedInReviewBoard90Warning,
                              deprecation_message):
            hook = HeaderActionHook(extension=self.extension, actions=[action])

        context = self._get_context()
        entries = hook.get_actions(context)
        self.assertEqual(len(entries), 1)
        self.assertEqual(entries[0].action_id, 'test')

        t = Template(
            '{% load actions %}'
            '{% actions_html "header" %}')

        content = t.render(context).strip()

        self.assertIn('id="%s-action"' % action['id'], content)

    def test_header_dropdown_action_hook(self) -> None:
        """Testing HeaderDropdownActionHook"""
        action = {
            'id': 'test-menu',
            'label': 'Test Menu',
            'items': [
                {
                    'id': 'test-action',
                    'image': 'test-image',
                    'image_height': 42,
                    'image_width': 42,
                    'label': 'Test Action',
                    'url': 'foo-url',
                }
            ]
        }

        deprecation_message = (
            'HeaderDropdownActionHook is deprecated and will be removed in '
            'Review Board 9. Your extension '
            '"reviewboard.extensions.tests.testcases.DummyExtension" '
            'will need to be updated to derive actions from '
            'reviewboard.actions.BaseAction and use ActionHook.'
        )

        with self.assertWarns(RemovedInReviewBoard90Warning,
                              deprecation_message):
            hook = HeaderDropdownActionHook(extension=self.extension,
                                            actions=[action])

        context = self._get_context()
        entries = hook.get_actions(context)
        self.assertEqual(len(entries), 2)
        self.assertEqual(entries[0].action_id, 'test-menu')
        self.assertEqual(entries[1].action_id, 'test-action')

        t = Template(
            '{% load actions %}'
            '{% actions_html "header" %}')

        content = t.render(context).strip()

        self.assertIn(('id="action-%s"' % action['id']), content)

        label = action['label']
        self.assertInHTML(
            f"""
            <a href="#" role="presentation" aria-label="{label}">
             {label}
             <span class="ink-i-dropdown"></span>
            </a>
            """,
            content)

    def _test_base_review_request_action_hook(
        self,
        url_name: str,
        hook_cls: Type[ActionHook],
        should_render: bool,
    ) -> None:
        """Test if the action hook renders or not at the given URL.

        Args:
            url_name (str):
                The name of the URL where each action is to be rendered.

            hook_cls (class):
                The class of the action hook to be tested.

            should_render (bool):
                The expected rendering behaviour.
        """
        hook = hook_cls(extension=self.extension, actions=[
            {
                'id': 'with-id',
                'label': 'Yes ID',
                'url': 'with-id-url',
            },
            {
                'label': 'No ID',
                'url': 'without-id-url',
            },
        ])

        try:
            context = self._get_context(url_name=url_name)

            entries = hook.get_actions(context)
            self.assertEqual(len(entries), 2)
            self.assertEqual(entries[0].action_id, 'with-id')
            self.assertEqual(entries[1].action_id, 'no-id')

            template = Template(
                '{% load actions %}'
                '{% actions_html "review-request" %}'
            )
            content = template.render(context)
            self.assertEqual(should_render, 'href="with-id-url"' in content)
            self.assertEqual(should_render, 'Yes ID' in content)
            self.assertEqual(should_render,
                             'id="no-id-action"' in content)
        finally:
            hook.disable_hook()

        content = template.render(context)
        self.assertNotIn('href="with-id-url"', content)
        self.assertNotIn('Yes ID', content)
        self.assertNotIn('id="action-no-id-dict"', content)

    def _test_review_request_dropdown_action_hook(
        self,
        url_name: str,
        hook_cls: Type[ActionHook],
    ) -> None:
        """Test if the dropdown action hook renders or not at the given URL.

        Args:
            url_name (str):
                The name of the URL where each action is to be rendered.

            hook_cls (class):
                The class of the dropdown action hook to be tested.
        """
        hook = hook_cls(extension=self.extension, actions=[
            {
                'id': 'test-menu-dict',
                'label': 'Menu Dict',
                'items': [
                    {
                        'id': 'with-id',
                        'label': 'Yes ID',
                        'url': 'with-id-url',
                    },
                    {
                        'label': 'No ID',
                        'url': 'without-id-url',
                    },
                ]
            },
        ])

        try:
            context = self._get_context(url_name=url_name)
            entries = hook.get_actions(context)
            self.assertEqual(len(entries), 3)
            self.assertEqual(entries[0].action_id, 'test-menu-dict')
            self.assertEqual(entries[1].action_id, 'with-id')
            self.assertEqual(entries[2].action_id, 'no-id')

            dropdown_icon_html = '<span class="ink-i-dropdown"></span>'

            template = Template(
                '{% load actions %}'
                '{% actions_html "review-request" %}'
            )
            content = template.render(context)

            self.assertIn('Yes ID', content)
            self.assertInHTML(
                f"""
                <a href="#" role="presentation" aria-label="Menu Dict">
                 Menu Dict
                 {dropdown_icon_html}
                </a>
                """,
                content)

            self.assertIn('id="action-test-menu-dict"', content)
            self.assertIn('href="with-id-url"', content)
            self.assertIn('id="no-id-action"', content)
        finally:
            hook.disable_hook()

        content = template.render(context)
        self.assertNotIn('Test Action', content)
        self.assertNotIn('Menu Instance', content)
        self.assertNotIn('id="test-menu-dict-action"', content)
        self.assertNotIn('href="with-id-url"', content)
        self.assertNotIn('id="no-id-dict-action"', content)

    def _build_action_template(
        self,
        action: Dict[str, Any],
    ) -> str:
        """Create HTML rendering of an action.

        Args:
            action (dict):
                Data about the action for rendering.

        Returns:
            str:
            HTML for the action rendering.
        """
        return ('<li><a id="%(id)s" href="%(url)s">'
                '<img src="%(image)s" width="%(image_width)s" '
                'height="%(image_height)s" border="0" alt="" />'
                '%(label)s</a></li>' % action)

    def _get_context(
        self,
        url_name: str = 'review-request-detail',
    ) -> Context:
        """Create a template rendering context.

        Args:
            url_name (str):
                The URL name to set on the request.

        Returns:
            django.template.Context:
            A rendering context to use for tests.
        """
        request = self.create_http_request(url_name=url_name)
        review_request = self.create_review_request(public=True)

        return Context({
            'request': request,
            'review_request': review_request,
            'perms': {
                'reviews': {
                    'can_change_status': True,
                    'can_edit_reviewrequest': True,
                    'delete_reviewrequest': True,
                },
            },
        })


class HideActionHookTests(BaseExtensionHookTestCase):
    """Tests for HideActionHook."""

    def test_hide_action_hook(self) -> None:
        """Testing HideActionHook"""
        HideActionHook(extension=self.extension,
                       action_ids=['support-menu'])

        action = actions_registry.get_action('support-menu')

        request = self.create_http_request()
        context = Context({
            'request': request,
        })

        self.assertFalse(action.should_render(context=context))



"""Rename ReviewRequest.last_review_timestamp.

This has been renamed to ``last_review_activity_timestamp``.

Version Added:
    1.6.15
"""

from django_evolution.mutations import RenameField


MUTATIONS = [
    RenameField('ReviewRequest', 'last_review_timestamp',
                'last_review_activity_timestamp',
                db_column='last_review_timestamp'),
]



"""Add an index to LocalSite.public.

This aids in generating LocalSite stats on larger installs.

Version Added:
    5.0
"""

from django_evolution.mutations import ChangeField


MUTATIONS = [
    ChangeField('LocalSite', 'public', db_index=True, initial=False),
]
